from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.tracing_utils import (
    add_http_request_source,
    has_span_streaming_enabled,
    propagate_trace_headers,
)
from sentry_sdk.utils import (
    SENSITIVE_DATA_SUBSTITUTE,
    capture_internal_exceptions,
    ensure_integration_enabled,
    parse_url,
)

if TYPE_CHECKING:
    from typing import Any

    from sentry_sdk._types import Attributes


try:
    from httpx import AsyncClient, Client, Request, Response
except ImportError:
    raise DidNotEnable("httpx is not installed")

__all__ = ["HttpxIntegration"]


class HttpxIntegration(Integration):
    identifier = "httpx"
    origin = f"auto.http.{identifier}"

    @staticmethod
    def setup_once() -> None:
        """
        httpx has its own transport layer and can be customized when needed,
        so patch Client.send and AsyncClient.send to support both synchronous and async interfaces.
        """
        _install_httpx_client()
        _install_httpx_async_client()


def _install_httpx_client() -> None:
    real_send = Client.send

    @ensure_integration_enabled(HttpxIntegration, real_send)
    def send(self: "Client", request: "Request", **kwargs: "Any") -> "Response":
        client = sentry_sdk.get_client()
        is_span_streaming_enabled = has_span_streaming_enabled(client.options)

        parsed_url = None
        with capture_internal_exceptions():
            parsed_url = parse_url(str(request.url), sanitize=False)

        if is_span_streaming_enabled:
            if sentry_sdk.traces.get_current_span() is None:
                propagate_trace_headers(client, request)
                return real_send(self, request, **kwargs)

            with sentry_sdk.traces.start_span(
                name="%s %s"
                % (
                    request.method,
                    parsed_url.url if parsed_url else SENSITIVE_DATA_SUBSTITUTE,
                ),
                attributes={
                    "sentry.op": OP.HTTP_CLIENT,
                    "sentry.origin": HttpxIntegration.origin,
                    "http.request.method": request.method,
                },
            ) as streamed_span:
                attributes: "Attributes" = {}

                if parsed_url is not None and should_send_default_pii():
                    attributes["url.full"] = parsed_url.url
                    if parsed_url.query:
                        attributes["url.query"] = parsed_url.query
                    if parsed_url.fragment:
                        attributes["url.fragment"] = parsed_url.fragment

                propagate_trace_headers(client, request)

                try:
                    rv = real_send(self, request, **kwargs)

                    streamed_span.status = "error" if rv.status_code >= 400 else "ok"
                    attributes["http.response.status_code"] = rv.status_code
                finally:
                    streamed_span.set_attributes(attributes)

                # Needs to happen within the context manager as we want to attach the
                # final data before the span finishes and is sent for ingesting.
                with capture_internal_exceptions():
                    add_http_request_source(streamed_span)
        else:
            with sentry_sdk.start_span(
                op=OP.HTTP_CLIENT,
                name="%s %s"
                % (
                    request.method,
                    parsed_url.url if parsed_url else SENSITIVE_DATA_SUBSTITUTE,
                ),
                origin=HttpxIntegration.origin,
            ) as span:
                span.set_data(SPANDATA.HTTP_METHOD, request.method)
                if parsed_url is not None:
                    span.set_data("url", parsed_url.url)
                    span.set_data(SPANDATA.HTTP_QUERY, parsed_url.query)
                    span.set_data(SPANDATA.HTTP_FRAGMENT, parsed_url.fragment)

                propagate_trace_headers(client, request)

                rv = real_send(self, request, **kwargs)

                span.set_http_status(rv.status_code)
                span.set_data("reason", rv.reason_phrase)

            with capture_internal_exceptions():
                add_http_request_source(span)

        return rv

    Client.send = send  # type: ignore


def _install_httpx_async_client() -> None:
    real_send = AsyncClient.send

    async def send(
        self: "AsyncClient", request: "Request", **kwargs: "Any"
    ) -> "Response":
        client = sentry_sdk.get_client()
        if client.get_integration(HttpxIntegration) is None:
            return await real_send(self, request, **kwargs)

        is_span_streaming_enabled = has_span_streaming_enabled(client.options)
        parsed_url = None
        with capture_internal_exceptions():
            parsed_url = parse_url(str(request.url), sanitize=False)

        if is_span_streaming_enabled:
            if sentry_sdk.traces.get_current_span() is None:
                propagate_trace_headers(client, request)
                return await real_send(self, request, **kwargs)

            with sentry_sdk.traces.start_span(
                name="%s %s"
                % (
                    request.method,
                    parsed_url.url if parsed_url else SENSITIVE_DATA_SUBSTITUTE,
                ),
                attributes={
                    "sentry.op": OP.HTTP_CLIENT,
                    "sentry.origin": HttpxIntegration.origin,
                    "http.request.method": request.method,
                },
            ) as streamed_span:
                attributes: "Attributes" = {}

                if parsed_url is not None and should_send_default_pii():
                    attributes["url.full"] = parsed_url.url
                    if parsed_url.query:
                        attributes["url.query"] = parsed_url.query
                    if parsed_url.fragment:
                        attributes["url.fragment"] = parsed_url.fragment

                propagate_trace_headers(client, request)

                try:
                    rv = await real_send(self, request, **kwargs)

                    streamed_span.status = "error" if rv.status_code >= 400 else "ok"
                    attributes["http.response.status_code"] = rv.status_code
                finally:
                    streamed_span.set_attributes(attributes)

                # Needs to happen within the context manager as we want to attach the
                # final data before the span finishes and is sent for ingesting.
                with capture_internal_exceptions():
                    add_http_request_source(streamed_span)
        else:
            with sentry_sdk.start_span(
                op=OP.HTTP_CLIENT,
                name="%s %s"
                % (
                    request.method,
                    parsed_url.url if parsed_url else SENSITIVE_DATA_SUBSTITUTE,
                ),
                origin=HttpxIntegration.origin,
            ) as span:
                span.set_data(SPANDATA.HTTP_METHOD, request.method)
                if parsed_url is not None:
                    span.set_data("url", parsed_url.url)
                    span.set_data(SPANDATA.HTTP_QUERY, parsed_url.query)
                    span.set_data(SPANDATA.HTTP_FRAGMENT, parsed_url.fragment)

                propagate_trace_headers(client, request)

                rv = await real_send(self, request, **kwargs)

                span.set_http_status(rv.status_code)
                span.set_data("reason", rv.reason_phrase)

            with capture_internal_exceptions():
                add_http_request_source(span)

        return rv

    AsyncClient.send = send  # type: ignore



from importlib import import_module

import sentry_sdk
from sentry_sdk import capture_event, get_client
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.integrations._wsgi_common import request_body_within_bounds
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.utils import (
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    package_version,
)

try:
    # importing like this is necessary due to name shadowing in ariadne
    # (ariadne.graphql is also a function)
    ariadne_graphql = import_module("ariadne.graphql")
except ImportError:
    raise DidNotEnable("ariadne is not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional

    from ariadne.types import (  # type: ignore
        GraphQLError,
        GraphQLResult,
        GraphQLSchema,
        QueryParser,
    )
    from graphql.language.ast import DocumentNode

    from sentry_sdk._types import Event, EventProcessor


class AriadneIntegration(Integration):
    identifier = "ariadne"

    @staticmethod
    def setup_once() -> None:
        version = package_version("ariadne")
        _check_minimum_version(AriadneIntegration, version)

        ignore_logger("ariadne")

        _patch_graphql()


def _patch_graphql() -> None:
    old_parse_query = ariadne_graphql.parse_query
    old_handle_errors = ariadne_graphql.handle_graphql_errors
    old_handle_query_result = ariadne_graphql.handle_query_result

    @ensure_integration_enabled(AriadneIntegration, old_parse_query)
    def _sentry_patched_parse_query(
        context_value: "Optional[Any]",
        query_parser: "Optional[QueryParser]",
        data: "Any",
    ) -> "DocumentNode":
        event_processor = _make_request_event_processor(data)
        sentry_sdk.get_isolation_scope().add_event_processor(event_processor)

        result = old_parse_query(context_value, query_parser, data)
        return result

    @ensure_integration_enabled(AriadneIntegration, old_handle_errors)
    def _sentry_patched_handle_graphql_errors(
        errors: "List[GraphQLError]", *args: "Any", **kwargs: "Any"
    ) -> "GraphQLResult":
        result = old_handle_errors(errors, *args, **kwargs)

        event_processor = _make_response_event_processor(result[1])
        sentry_sdk.get_isolation_scope().add_event_processor(event_processor)

        client = get_client()
        if client.is_active():
            with capture_internal_exceptions():
                for error in errors:
                    event, hint = event_from_exception(
                        error,
                        client_options=client.options,
                        mechanism={
                            "type": AriadneIntegration.identifier,
                            "handled": False,
                        },
                    )
                    capture_event(event, hint=hint)

        return result

    @ensure_integration_enabled(AriadneIntegration, old_handle_query_result)
    def _sentry_patched_handle_query_result(
        result: "Any", *args: "Any", **kwargs: "Any"
    ) -> "GraphQLResult":
        query_result = old_handle_query_result(result, *args, **kwargs)

        event_processor = _make_response_event_processor(query_result[1])
        sentry_sdk.get_isolation_scope().add_event_processor(event_processor)

        client = get_client()
        if client.is_active():
            with capture_internal_exceptions():
                for error in result.errors or []:
                    event, hint = event_from_exception(
                        error,
                        client_options=client.options,
                        mechanism={
                            "type": AriadneIntegration.identifier,
                            "handled": False,
                        },
                    )
                    capture_event(event, hint=hint)

        return query_result

    ariadne_graphql.parse_query = _sentry_patched_parse_query  # type: ignore
    ariadne_graphql.handle_graphql_errors = _sentry_patched_handle_graphql_errors  # type: ignore
    ariadne_graphql.handle_query_result = _sentry_patched_handle_query_result  # type: ignore


def _make_request_event_processor(data: "GraphQLSchema") -> "EventProcessor":
    """Add request data and api_target to events."""

    def inner(event: "Event", hint: "dict[str, Any]") -> "Event":
        if not isinstance(data, dict):
            return event

        with capture_internal_exceptions():
            try:
                content_length = int(
                    (data.get("headers") or {}).get("Content-Length", 0)
                )
            except (TypeError, ValueError):
                return event

            if should_send_default_pii() and request_body_within_bounds(
                get_client(), content_length
            ):
                request_info = event.setdefault("request", {})
                request_info["api_target"] = "graphql"
                request_info["data"] = data

            elif event.get("request", {}).get("data"):
                del event["request"]["data"]

        return event

    return inner


def _make_response_event_processor(response: "Dict[str, Any]") -> "EventProcessor":
    """Add response data to the event's response context."""

    def inner(event: "Event", hint: "dict[str, Any]") -> "Event":
        with capture_internal_exceptions():
            if should_send_default_pii() and response.get("errors"):
                contexts = event.setdefault("contexts", {})
                contexts["response"] = {
                    "data": response,
                }

        return event

    return inner



import sys
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.integrations import Integration
from sentry_sdk.scope import add_global_event_processor

if TYPE_CHECKING:
    from typing import Optional

    from sentry_sdk._types import Event, Hint


class ArgvIntegration(Integration):
    identifier = "argv"

    @staticmethod
    def setup_once() -> None:
        @add_global_event_processor
        def processor(event: "Event", hint: "Optional[Hint]") -> "Optional[Event]":
            if sentry_sdk.get_client().get_integration(ArgvIntegration) is not None:
                extra = event.setdefault("extra", {})
                # If some event processor decided to set extra to e.g. an
                # `int`, don't crash. Not here.
                if isinstance(extra, dict):
                    extra["sys.argv"] = sys.argv

            return event



import functools
import json
import re
import sys
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from os import environ
from typing import TYPE_CHECKING
from urllib.parse import urlencode

import sentry_sdk
from sentry_sdk.api import continue_trace
from sentry_sdk.consts import OP
from sentry_sdk.integrations import Integration
from sentry_sdk.integrations._wsgi_common import _filter_headers
from sentry_sdk.integrations.cloud_resource_context import (
    CLOUD_PLATFORM,
    CLOUD_PROVIDER,
)
from sentry_sdk.scope import Scope, should_send_default_pii
from sentry_sdk.traces import SegmentSource
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    AnnotatedValue,
    TimeoutThread,
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    logger,
    reraise,
)

if TYPE_CHECKING:
    from typing import Any, Callable, Optional, TypeVar

    from sentry_sdk._types import Event, EventProcessor, Hint

    F = TypeVar("F", bound=Callable[..., Any])

# Constants
TIMEOUT_WARNING_BUFFER = 1500  # Buffer time required to send timeout warning to Sentry
MILLIS_TO_SECONDS = 1000.0


def _wrap_init_error(init_error: "F") -> "F":
    @ensure_integration_enabled(AwsLambdaIntegration, init_error)
    def sentry_init_error(*args: "Any", **kwargs: "Any") -> "Any":
        client = sentry_sdk.get_client()

        with capture_internal_exceptions():
            sentry_sdk.get_isolation_scope().clear_breadcrumbs()

            exc_info = sys.exc_info()
            if exc_info and all(exc_info):
                sentry_event, hint = event_from_exception(
                    exc_info,
                    client_options=client.options,
                    mechanism={"type": "aws_lambda", "handled": False},
                )
                sentry_sdk.capture_event(sentry_event, hint=hint)

            else:
                # Fall back to AWS lambdas JSON representation of the error
                error_info = args[1]
                if isinstance(error_info, str):
                    error_info = json.loads(error_info)
                sentry_event = _event_from_error_json(error_info)
                sentry_sdk.capture_event(sentry_event)

        return init_error(*args, **kwargs)

    return sentry_init_error  # type: ignore


def _wrap_handler(handler: "F") -> "F":
    @functools.wraps(handler)
    def sentry_handler(
        aws_event: "Any", aws_context: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        # Per https://docs.aws.amazon.com/lambda/latest/dg/python-handler.html,
        # `event` here is *likely* a dictionary, but also might be a number of
        # other types (str, int, float, None).
        #
        # In some cases, it is a list (if the user is batch-invoking their
        # function, for example), in which case we'll use the first entry as a
        # representative from which to try pulling request data. (Presumably it
        # will be the same for all events in the list, since they're all hitting
        # the lambda in the same request.)

        client = sentry_sdk.get_client()
        integration = client.get_integration(AwsLambdaIntegration)

        if integration is None:
            return handler(aws_event, aws_context, *args, **kwargs)

        if isinstance(aws_event, list) and len(aws_event) >= 1:
            request_data = aws_event[0]
            batch_size = len(aws_event)
        else:
            request_data = aws_event
            batch_size = 1

        if not isinstance(request_data, dict):
            # If we're not dealing with a dictionary, we won't be able to get
            # headers, path, http method, etc in any case, so it's fine that
            # this is empty
            request_data = {}

        configured_time = aws_context.get_remaining_time_in_millis()
        aws_region = aws_context.invoked_function_arn.split(":")[3]

        with sentry_sdk.isolation_scope() as scope:
            timeout_thread = None
            with capture_internal_exceptions():
                scope.clear_breadcrumbs()
                scope.add_event_processor(
                    _make_request_event_processor(
                        request_data, aws_context, configured_time
                    )
                )
                scope.set_tag("aws_region", aws_region)
                if batch_size > 1:
                    scope.set_tag("batch_request", True)
                    scope.set_tag("batch_size", batch_size)

                # Starting the Timeout thread only if the configured time is greater than Timeout warning
                # buffer and timeout_warning parameter is set True.
                if (
                    integration.timeout_warning
                    and configured_time > TIMEOUT_WARNING_BUFFER
                ):
                    waiting_time = (
                        configured_time - TIMEOUT_WARNING_BUFFER
                    ) / MILLIS_TO_SECONDS

                    timeout_thread = TimeoutThread(
                        waiting_time,
                        configured_time / MILLIS_TO_SECONDS,
                        isolation_scope=scope,
                        current_scope=sentry_sdk.get_current_scope(),
                    )

                    # Starting the thread to raise timeout warning exception
                    timeout_thread.start()

            headers = request_data.get("headers", {})
            # Some AWS Services (ie. EventBridge) set headers as a list
            # or None, so we must ensure it is a dict
            if not isinstance(headers, dict):
                headers = {}

            header_attributes: "dict[str, Any]" = {}
            for header, header_value in _filter_headers(
                headers, use_annotated_value=False
            ).items():
                header_attributes[f"http.request.header.{header.lower()}"] = (
                    header_value
                )

            additional_attributes: "dict[str, Any]" = {}
            if "httpMethod" in request_data:
                additional_attributes["http.request.method"] = request_data[
                    "httpMethod"
                ]

            if should_send_default_pii() and "queryStringParameters" in request_data:
                qs = request_data["queryStringParameters"]
                if qs:
                    additional_attributes["url.query"] = urlencode(qs)

            sampling_context = {
                "aws_event": aws_event,
                "aws_context": aws_context,
            }

            function_name = aws_context.function_name

            if has_span_streaming_enabled(client.options):
                sentry_sdk.traces.continue_trace(headers)
                Scope.set_custom_sampling_context(sampling_context)
                span_ctx = sentry_sdk.traces.start_span(
                    name=function_name,
                    parent_span=None,
                    attributes={
                        "sentry.op": OP.FUNCTION_AWS,
                        "sentry.origin": AwsLambdaIntegration.origin,
                        "sentry.span.source": SegmentSource.COMPONENT,
                        "cloud.region": aws_region,
                        "cloud.resource_id": aws_context.invoked_function_arn,
                        "cloud.platform": CLOUD_PLATFORM.AWS_LAMBDA,
                        "cloud.provider": CLOUD_PROVIDER.AWS,
                        "faas.name": function_name,
                        "faas.invocation_id": aws_context.aws_request_id,
                        "faas.version": aws_context.function_version,
                        "aws.lambda.invoked_arn": aws_context.invoked_function_arn,
                        "aws.log.group.names": [aws_context.log_group_name],
                        "aws.log.stream.names": [aws_context.log_stream_name],
                        "messaging.batch.message_count": batch_size,
                        **header_attributes,
                        **additional_attributes,
                    },
                )
            else:
                transaction = continue_trace(
                    headers,
                    op=OP.FUNCTION_AWS,
                    name=function_name,
                    source=TransactionSource.COMPONENT,
                    origin=AwsLambdaIntegration.origin,
                )

                span_ctx = sentry_sdk.start_transaction(
                    transaction, custom_sampling_context=sampling_context
                )

            with span_ctx:
                try:
                    return handler(aws_event, aws_context, *args, **kwargs)
                except Exception:
                    exc_info = sys.exc_info()
                    sentry_event, hint = event_from_exception(
                        exc_info,
                        client_options=client.options,
                        mechanism={"type": "aws_lambda", "handled": False},
                    )
                    sentry_sdk.capture_event(sentry_event, hint=hint)
                    reraise(*exc_info)
                finally:
                    if timeout_thread:
                        timeout_thread.stop()

    return sentry_handler  # type: ignore


def _drain_queue() -> None:
    with capture_internal_exceptions():
        client = sentry_sdk.get_client()
        integration = client.get_integration(AwsLambdaIntegration)
        if integration is not None:
            # Flush out the event queue before AWS kills the
            # process.
            client.flush()


class AwsLambdaIntegration(Integration):
    identifier = "aws_lambda"
    origin = f"auto.function.{identifier}"

    def __init__(self, timeout_warning: bool = False) -> None:
        self.timeout_warning = timeout_warning

    @staticmethod
    def setup_once() -> None:
        lambda_bootstrap = get_lambda_bootstrap()
        if not lambda_bootstrap:
            logger.warning(
                "Not running in AWS Lambda environment, "
                "AwsLambdaIntegration disabled (could not find bootstrap module)"
            )
            return

        if not hasattr(lambda_bootstrap, "handle_event_request"):
            logger.warning(
                "Not running in AWS Lambda environment, "
                "AwsLambdaIntegration disabled (could not find handle_event_request)"
            )
            return

        pre_37 = hasattr(lambda_bootstrap, "handle_http_request")  # Python 3.6

        if pre_37:
            old_handle_event_request = lambda_bootstrap.handle_event_request

            def sentry_handle_event_request(
                request_handler: "Any", *args: "Any", **kwargs: "Any"
            ) -> "Any":
                request_handler = _wrap_handler(request_handler)
                return old_handle_event_request(request_handler, *args, **kwargs)

            lambda_bootstrap.handle_event_request = sentry_handle_event_request

            old_handle_http_request = lambda_bootstrap.handle_http_request

            def sentry_handle_http_request(
                request_handler: "Any", *args: "Any", **kwargs: "Any"
            ) -> "Any":
                request_handler = _wrap_handler(request_handler)
                return old_handle_http_request(request_handler, *args, **kwargs)

            lambda_bootstrap.handle_http_request = sentry_handle_http_request

            # Patch to_json to drain the queue. This should work even when the
            # SDK is initialized inside of the handler

            old_to_json = lambda_bootstrap.to_json

            def sentry_to_json(*args: "Any", **kwargs: "Any") -> "Any":
                _drain_queue()
                return old_to_json(*args, **kwargs)

            lambda_bootstrap.to_json = sentry_to_json
        else:
            lambda_bootstrap.LambdaRuntimeClient.post_init_error = _wrap_init_error(
                lambda_bootstrap.LambdaRuntimeClient.post_init_error
            )

            old_handle_event_request = lambda_bootstrap.handle_event_request

            def sentry_handle_event_request(  # type: ignore
                lambda_runtime_client, request_handler, *args, **kwargs
            ):
                request_handler = _wrap_handler(request_handler)
                return old_handle_event_request(
                    lambda_runtime_client, request_handler, *args, **kwargs
                )

            lambda_bootstrap.handle_event_request = sentry_handle_event_request

            # Patch the runtime client to drain the queue. This should work
            # even when the SDK is initialized inside of the handler

            def _wrap_post_function(f: "F") -> "F":
                def inner(*args: "Any", **kwargs: "Any") -> "Any":
                    _drain_queue()
                    return f(*args, **kwargs)

                return inner  # type: ignore

            lambda_bootstrap.LambdaRuntimeClient.post_invocation_result = (
                _wrap_post_function(
                    lambda_bootstrap.LambdaRuntimeClient.post_invocation_result
                )
            )
            lambda_bootstrap.LambdaRuntimeClient.post_invocation_error = (
                _wrap_post_function(
                    lambda_bootstrap.LambdaRuntimeClient.post_invocation_error
                )
            )


def get_lambda_bootstrap() -> "Optional[Any]":
    # Python 3.7: If the bootstrap module is *already imported*, it is the
    # one we actually want to use (no idea what's in __main__)
    #
    # Python 3.8: bootstrap is also importable, but will be the same file
    # as __main__ imported under a different name:
    #
    #     sys.modules['__main__'].__file__ == sys.modules['bootstrap'].__file__
    #     sys.modules['__main__'] is not sys.modules['bootstrap']
    #
    # Python 3.9: bootstrap is in __main__.awslambdaricmain
    #
    # On container builds using the `aws-lambda-python-runtime-interface-client`
    # (awslamdaric) module, bootstrap is located in sys.modules['__main__'].bootstrap
    #
    # Such a setup would then make all monkeypatches useless.
    if "bootstrap" in sys.modules:
        return sys.modules["bootstrap"]
    elif "__main__" in sys.modules:
        module = sys.modules["__main__"]
        # python3.9 runtime
        if hasattr(module, "awslambdaricmain") and hasattr(
            module.awslambdaricmain, "bootstrap"
        ):
            return module.awslambdaricmain.bootstrap
        elif hasattr(module, "bootstrap"):
            # awslambdaric python module in container builds
            return module.bootstrap

        # python3.8 runtime
        return module
    else:
        return None


def _make_request_event_processor(
    aws_event: "Any", aws_context: "Any", configured_timeout: "Any"
) -> "EventProcessor":
    start_time = datetime.now(timezone.utc)

    def event_processor(
        sentry_event: "Event", hint: "Hint", start_time: "datetime" = start_time
    ) -> "Optional[Event]":
        remaining_time_in_milis = aws_context.get_remaining_time_in_millis()
        exec_duration = configured_timeout - remaining_time_in_milis

        extra = sentry_event.setdefault("extra", {})
        extra["lambda"] = {
            "function_name": aws_context.function_name,
            "function_version": aws_context.function_version,
            "invoked_function_arn": aws_context.invoked_function_arn,
            "aws_request_id": aws_context.aws_request_id,
            "execution_duration_in_millis": exec_duration,
            "remaining_time_in_millis": remaining_time_in_milis,
        }

        extra["cloudwatch logs"] = {
            "url": _get_cloudwatch_logs_url(aws_context, start_time),
            "log_group": aws_context.log_group_name,
            "log_stream": aws_context.log_stream_name,
        }

        request = sentry_event.get("request", {})

        if "httpMethod" in aws_event:
            request["method"] = aws_event["httpMethod"]

        request["url"] = _get_url(aws_event, aws_context)

        if "queryStringParameters" in aws_event:
            request["query_string"] = aws_event["queryStringParameters"]

        if "headers" in aws_event:
            request["headers"] = _filter_headers(aws_event["headers"])

        if should_send_default_pii():
            user_info = sentry_event.setdefault("user", {})

            identity = aws_event.get("identity")
            if identity is None:
                identity = {}

            id = identity.get("userArn")
            if id is not None:
                user_info.setdefault("id", id)

            ip = identity.get("sourceIp")
            if ip is not None:
                user_info.setdefault("ip_address", ip)

            if "body" in aws_event:
                request["data"] = aws_event.get("body", "")
        else:
            if aws_event.get("body", None):
                # Unfortunately couldn't find a way to get structured body from AWS
                # event. Meaning every body is unstructured to us.
                request["data"] = AnnotatedValue.removed_because_raw_data()

        sentry_event["request"] = deepcopy(request)

        return sentry_event

    return event_processor


def _get_url(aws_event: "Any", aws_context: "Any") -> str:
    path = aws_event.get("path", None)

    headers = aws_event.get("headers")
    if headers is None:
        headers = {}

    host = headers.get("Host", None)
    proto = headers.get("X-Forwarded-Proto", None)
    if proto and host and path:
        return "{}://{}{}".format(proto, host, path)
    return "awslambda:///{}".format(aws_context.function_name)


def _get_cloudwatch_logs_url(aws_context: "Any", start_time: "datetime") -> str:
    """
    Generates a CloudWatchLogs console URL based on the context object

    Arguments:
        aws_context {Any} -- context from lambda handler

    Returns:
        str -- AWS Console URL to logs.
    """
    formatstring = "%Y-%m-%dT%H:%M:%SZ"
    region = environ.get("AWS_REGION", "")

    url = (
        "https://console.{domain}/cloudwatch/home?region={region}"
        "#logEventViewer:group={log_group};stream={log_stream}"
        ";start={start_time};end={end_time}"
    ).format(
        domain="amazonaws.cn" if region.startswith("cn-") else "aws.amazon.com",
        region=region,
        log_group=aws_context.log_group_name,
        log_stream=aws_context.log_stream_name,
        start_time=(start_time - timedelta(seconds=1)).strftime(formatstring),
        end_time=(datetime.now(timezone.utc) + timedelta(seconds=2)).strftime(
            formatstring
        ),
    )

    return url


def _parse_formatted_traceback(formatted_tb: "list[str]") -> "list[dict[str, Any]]":
    frames = []
    for frame in formatted_tb:
        match = re.match(r'File "(.+)", line (\d+), in (.+)', frame.strip())
        if match:
            file_name, line_number, func_name = match.groups()
            line_number = int(line_number)
            frames.append(
                {
                    "filename": file_name,
                    "function": func_name,
                    "lineno": line_number,
                    "vars": None,
                    "pre_context": None,
                    "context_line": None,
                    "post_context": None,
                }
            )
    return frames


def _event_from_error_json(error_json: "dict[str, Any]") -> "Event":
    """
    Converts the error JSON from AWS Lambda into a Sentry error event.
    This is not a full fletched event, but better than nothing.

    This is an example of where AWS creates the error JSON:
    https://github.com/aws/aws-lambda-python-runtime-interface-client/blob/2.2.1/awslambdaric/bootstrap.py#L479
    """
    event: "Event" = {
        "level": "error",
        "exception": {
            "values": [
                {
                    "type": error_json.get("errorType"),
                    "value": error_json.get("errorMessage"),
                    "stacktrace": {
                        "frames": _parse_formatted_traceback(
                            error_json.get("stackTrace", [])
                        ),
                    },
                    "mechanism": {
                        "type": "aws_lambda",
                        "handled": False,
                    },
                }
            ],
        },
    }

    return event



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union

    from openai.types.responses import ResponseInputItemParam, ResponseInputParam


def _is_system_instruction(message: "ResponseInputItemParam") -> bool:
    if not isinstance(message, dict) or not message.get("role") == "system":
        return False

    return "type" not in message or message["type"] == "message"


def _get_system_instructions(
    messages: "Union[str, ResponseInputParam]",
) -> "list[ResponseInputItemParam]":
    if not isinstance(messages, list):
        return []

    return [message for message in messages if _is_system_instruction(message)]



"""
Instrumentation for Redis

https://github.com/redis/redis-py
"""

from typing import TYPE_CHECKING

from sentry_sdk.integrations.redis._sync_common import (
    patch_redis_client,
    patch_redis_pipeline,
)
from sentry_sdk.integrations.redis.modules.queries import _set_db_data

if TYPE_CHECKING:
    from typing import Any, Sequence


def _get_redis_command_args(command: "Any") -> "Sequence[Any]":
    return command[0]


def _patch_redis(StrictRedis: "Any", client: "Any") -> None:  # noqa: N803
    patch_redis_client(
        StrictRedis,
        is_cluster=False,
        set_db_data_fn=_set_db_data,
    )
    patch_redis_pipeline(
        client.Pipeline,
        is_cluster=False,
        get_command_args_fn=_get_redis_command_args,
        set_db_data_fn=_set_db_data,
    )
    try:
        strict_pipeline = client.StrictPipeline
    except AttributeError:
        pass
    else:
        patch_redis_pipeline(
            strict_pipeline,
            is_cluster=False,
            get_command_args_fn=_get_redis_command_args,
            set_db_data_fn=_set_db_data,
        )

    try:
        import redis.asyncio
    except ImportError:
        pass
    else:
        from sentry_sdk.integrations.redis._async_common import (
            patch_redis_async_client,
            patch_redis_async_pipeline,
        )

        patch_redis_async_client(
            redis.asyncio.client.StrictRedis,
            is_cluster=False,
            set_db_data_fn=_set_db_data,
        )
        patch_redis_async_pipeline(
            redis.asyncio.client.Pipeline,
            False,
            _get_redis_command_args,
            set_db_data_fn=_set_db_data,
        )



import json
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.ai._openai_completions_api import _transform_system_instructions
from sentry_sdk.ai._openai_responses_api import (
    _get_system_instructions,
    _is_system_instruction,
)
from sentry_sdk.ai.utils import (
    GEN_AI_ALLOWED_MESSAGE_ROLES,
    normalize_message_role,
    normalize_message_roles,
    set_data_normalized,
    truncate_and_annotate_messages,
)
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import should_truncate_gen_ai_input
from sentry_sdk.utils import event_from_exception, safe_serialize

if TYPE_CHECKING:
    from typing import Any, Union

    from agents import TResponseInputItem, Usage

    from sentry_sdk._types import TextPart

try:
    import agents

except ImportError:
    raise DidNotEnable("OpenAI Agents not installed")


def _capture_exception(exc: "Any") -> None:
    event, hint = event_from_exception(
        exc,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": "openai_agents", "handled": False},
    )
    sentry_sdk.capture_event(event, hint=hint)


def _set_agent_data(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]", agent: "agents.Agent"
) -> None:
    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )

    set_on_span(
        SPANDATA.GEN_AI_SYSTEM, "openai"
    )  # See footnote for  https://opentelemetry.io/docs/specs/semconv/registry/attributes/gen-ai/#gen-ai-system for explanation why.

    set_on_span(SPANDATA.GEN_AI_AGENT_NAME, agent.name)

    if agent.model_settings.max_tokens:
        set_on_span(SPANDATA.GEN_AI_REQUEST_MAX_TOKENS, agent.model_settings.max_tokens)

    # Get model name from agent.model or fall back to request model (for when agent.model is None/default)
    model_name = None
    if agent.model:
        model_name = agent.model.model if hasattr(agent.model, "model") else agent.model
    elif hasattr(agent, "_sentry_request_model"):
        model_name = agent._sentry_request_model

    if model_name:
        set_on_span(SPANDATA.GEN_AI_REQUEST_MODEL, model_name)

    if agent.model_settings.presence_penalty:
        set_on_span(
            SPANDATA.GEN_AI_REQUEST_PRESENCE_PENALTY,
            agent.model_settings.presence_penalty,
        )

    if agent.model_settings.temperature:
        set_on_span(
            SPANDATA.GEN_AI_REQUEST_TEMPERATURE, agent.model_settings.temperature
        )

    if agent.model_settings.top_p:
        set_on_span(SPANDATA.GEN_AI_REQUEST_TOP_P, agent.model_settings.top_p)

    if agent.model_settings.frequency_penalty:
        set_on_span(
            SPANDATA.GEN_AI_REQUEST_FREQUENCY_PENALTY,
            agent.model_settings.frequency_penalty,
        )

    if len(agent.tools) > 0:
        set_on_span(
            SPANDATA.GEN_AI_REQUEST_AVAILABLE_TOOLS,
            safe_serialize([vars(tool) for tool in agent.tools]),
        )


def _set_usage_data(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]", usage: "Usage"
) -> None:
    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )
    set_on_span(SPANDATA.GEN_AI_USAGE_INPUT_TOKENS, usage.input_tokens)
    set_on_span(
        SPANDATA.GEN_AI_USAGE_INPUT_TOKENS_CACHED,
        usage.input_tokens_details.cached_tokens,
    )
    set_on_span(SPANDATA.GEN_AI_USAGE_OUTPUT_TOKENS, usage.output_tokens)
    set_on_span(
        SPANDATA.GEN_AI_USAGE_OUTPUT_TOKENS_REASONING,
        usage.output_tokens_details.reasoning_tokens,
    )
    set_on_span(SPANDATA.GEN_AI_USAGE_TOTAL_TOKENS, usage.total_tokens)


def _set_input_data(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]",
    get_response_kwargs: "dict[str, Any]",
) -> None:
    if not should_send_default_pii():
        return
    request_messages = []

    messages: "str | list[TResponseInputItem]" = get_response_kwargs.get("input", [])

    instructions_text_parts: "list[TextPart]" = []
    explicit_instructions = get_response_kwargs.get("system_instructions")
    if explicit_instructions is not None:
        instructions_text_parts.append(
            {
                "type": "text",
                "content": explicit_instructions,
            }
        )

    system_instructions = _get_system_instructions(messages)

    # Deliberate use of function accepting completions API type because
    # of shared structure FOR THIS PURPOSE ONLY.
    instructions_text_parts += _transform_system_instructions(system_instructions)

    if len(instructions_text_parts) > 0:
        if isinstance(span, StreamedSpan):
            span.set_attribute(
                SPANDATA.GEN_AI_SYSTEM_INSTRUCTIONS,
                json.dumps(instructions_text_parts),
            )
        else:
            span.set_data(
                SPANDATA.GEN_AI_SYSTEM_INSTRUCTIONS,
                json.dumps(instructions_text_parts),
            )

    non_system_messages = [
        message for message in messages if not _is_system_instruction(message)
    ]
    for message in non_system_messages:
        if "role" in message:
            normalized_role = normalize_message_role(message.get("role"))  # type: ignore
            content = message.get("content")  # type: ignore
            request_messages.append(
                {
                    "role": normalized_role,
                    "content": (
                        [{"type": "text", "text": content}]
                        if isinstance(content, str)
                        else content
                    ),
                }
            )
        else:
            if message.get("type") == "function_call":  # type: ignore
                request_messages.append(
                    {
                        "role": GEN_AI_ALLOWED_MESSAGE_ROLES.ASSISTANT,
                        "content": [message],
                    }
                )
            elif message.get("type") == "function_call_output":  # type: ignore
                request_messages.append(
                    {
                        "role": GEN_AI_ALLOWED_MESSAGE_ROLES.TOOL,
                        "content": [message],
                    }
                )

    normalized_messages = normalize_message_roles(request_messages)
    client = sentry_sdk.get_client()
    scope = sentry_sdk.get_current_scope()
    messages_data = (
        truncate_and_annotate_messages(normalized_messages, span, scope)
        if should_truncate_gen_ai_input(client.options)
        else normalized_messages
    )
    if messages_data is not None:
        set_data_normalized(
            span,
            SPANDATA.GEN_AI_REQUEST_MESSAGES,
            messages_data,
            unpack=False,
        )


def _set_output_data(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]", result: "Any"
) -> None:
    if not should_send_default_pii():
        return

    output_messages: "dict[str, list[Any]]" = {
        "response": [],
        "tool": [],
    }

    for output in result.output:
        if output.type == "function_call":
            output_messages["tool"].append(output.dict())
        elif output.type == "message":
            for output_message in output.content:
                try:
                    output_messages["response"].append(output_message.text)
                except AttributeError:
                    # Unknown output message type, just return the json
                    output_messages["response"].append(output_message.dict())

    if len(output_messages["tool"]) > 0:
        if isinstance(span, StreamedSpan):
            span.set_attribute(
                SPANDATA.GEN_AI_RESPONSE_TOOL_CALLS,
                safe_serialize(output_messages["tool"]),
            )
        else:
            span.set_data(
                SPANDATA.GEN_AI_RESPONSE_TOOL_CALLS,
                safe_serialize(output_messages["tool"]),
            )

    if len(output_messages["response"]) > 0:
        set_data_normalized(
            span, SPANDATA.GEN_AI_RESPONSE_TEXT, output_messages["response"]
        )



import sys
from datetime import datetime
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.api import continue_trace, get_baggage, get_traceparent
from sentry_sdk.consts import OP, SPANDATA, SPANSTATUS
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import SegmentSource, SpanStatus, StreamedSpan
from sentry_sdk.tracing import (
    BAGGAGE_HEADER_NAME,
    SENTRY_TRACE_HEADER_NAME,
    TransactionSource,
)
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    SENSITIVE_DATA_SUBSTITUTE,
    _register_control_flow_exception,
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    reraise,
)

if TYPE_CHECKING:
    from typing import Any, Callable, Optional, TypeVar, Union

    from sentry_sdk._types import Event, EventProcessor, Hint
    from sentry_sdk.utils import ExcInfo

    F = TypeVar("F", bound=Callable[..., Any])

try:
    from huey.api import Huey, PeriodicTask, Result, ResultGroup, Task
    from huey.exceptions import CancelExecution, RetryTask, TaskLockedException
except ImportError:
    raise DidNotEnable("Huey is not installed")

try:
    from huey.api import chord as HueyChord
    from huey.api import group as HueyGroup
except ImportError:
    HueyChord = None
    HueyGroup = None


HUEY_CONTROL_FLOW_EXCEPTIONS = (CancelExecution, RetryTask, TaskLockedException)


class HueyIntegration(Integration):
    identifier = "huey"
    origin = f"auto.queue.{identifier}"

    @staticmethod
    def setup_once() -> None:
        patch_enqueue()
        patch_execute()
        _register_control_flow_exception(
            [CancelExecution, RetryTask, TaskLockedException]
        )


def patch_enqueue() -> None:
    old_enqueue = Huey.enqueue

    @ensure_integration_enabled(HueyIntegration, old_enqueue)
    def _sentry_enqueue(
        self: "Huey", item: "Any"
    ) -> "Optional[Union[Result, ResultGroup]]":
        if HueyChord is not None and isinstance(item, HueyChord):
            span_name = "Huey Chord"
        elif HueyGroup is not None and isinstance(item, HueyGroup):
            span_name = "Huey Task Group"
        else:
            span_name = item.name

        is_span_streaming_enabled = has_span_streaming_enabled(
            sentry_sdk.get_client().options
        )

        no_headers_types = (PeriodicTask,) + tuple(
            t for t in [HueyGroup, HueyChord] if t is not None
        )

        if is_span_streaming_enabled and sentry_sdk.traces.get_current_span() is None:
            if not isinstance(item, no_headers_types):
                item.kwargs["sentry_headers"] = {
                    BAGGAGE_HEADER_NAME: get_baggage(),
                    SENTRY_TRACE_HEADER_NAME: get_traceparent(),
                }
            return old_enqueue(self, item)

        if is_span_streaming_enabled:
            span_ctx = sentry_sdk.traces.start_span(
                name=span_name,
                attributes={
                    "sentry.op": OP.QUEUE_SUBMIT_HUEY,
                    "sentry.origin": HueyIntegration.origin,
                    SPANDATA.MESSAGING_DESTINATION_NAME: self.name,
                },
            )
        else:
            span_ctx = sentry_sdk.start_span(
                op=OP.QUEUE_SUBMIT_HUEY,
                name=span_name,
                origin=HueyIntegration.origin,
            )
            span_ctx.set_data(SPANDATA.MESSAGING_DESTINATION_NAME, self.name)

        with span_ctx:
            if not isinstance(item, no_headers_types):
                # Attach trace propagation data to task kwargs. We do
                # not do this for periodic tasks, as these don't
                # really have an originating transaction.
                # Additionally, we do not do this for Huey groups or chords, as enqueue will
                # recursively call this method for each task within the list, resulting
                # in the trace propagation data being attached to each task individually
                # (which we want)
                item.kwargs["sentry_headers"] = {
                    BAGGAGE_HEADER_NAME: get_baggage(),
                    SENTRY_TRACE_HEADER_NAME: get_traceparent(),
                }
            return old_enqueue(self, item)

    Huey.enqueue = _sentry_enqueue


def _make_event_processor(task: "Any") -> "EventProcessor":
    def event_processor(event: "Event", hint: "Hint") -> "Optional[Event]":
        with capture_internal_exceptions():
            tags = event.setdefault("tags", {})
            tags["huey_task_id"] = task.id
            tags["huey_task_retry"] = task.default_retries > task.retries
            extra = event.setdefault("extra", {})
            extra["huey-job"] = {
                "task": task.name,
                "args": (
                    task.args
                    if should_send_default_pii()
                    else SENSITIVE_DATA_SUBSTITUTE
                ),
                "kwargs": (
                    task.kwargs
                    if should_send_default_pii()
                    else SENSITIVE_DATA_SUBSTITUTE
                ),
                "retry": (task.default_retries or 0) - task.retries,
            }

        return event

    return event_processor


def _capture_exception(exc_info: "ExcInfo") -> None:
    scope = sentry_sdk.get_current_scope()
    is_span_streaming_enabled = has_span_streaming_enabled(
        sentry_sdk.get_client().options
    )

    if exc_info[0] in HUEY_CONTROL_FLOW_EXCEPTIONS:
        if not is_span_streaming_enabled:
            scope.transaction.set_status(SPANSTATUS.ABORTED)
        elif type(scope._span) is StreamedSpan:
            scope._span._segment.status = SpanStatus.OK
        return

    if not is_span_streaming_enabled:
        scope.transaction.set_status(SPANSTATUS.INTERNAL_ERROR)
    elif type(scope._span) is StreamedSpan:
        scope._span._segment.status = SpanStatus.ERROR

    event, hint = event_from_exception(
        exc_info,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": HueyIntegration.identifier, "handled": False},
    )
    scope.capture_event(event, hint=hint)


def _wrap_task_execute(func: "F") -> "F":
    @ensure_integration_enabled(HueyIntegration, func)
    def _sentry_execute(*args: "Any", **kwargs: "Any") -> "Any":
        try:
            result = func(*args, **kwargs)
        except Exception:
            exc_info = sys.exc_info()
            _capture_exception(exc_info)
            reraise(*exc_info)

        return result

    return _sentry_execute  # type: ignore


def patch_execute() -> None:
    old_execute = Huey._execute

    @ensure_integration_enabled(HueyIntegration, old_execute)
    def _sentry_execute(
        self: "Huey", task: "Task", timestamp: "Optional[datetime]" = None
    ) -> "Any":
        with sentry_sdk.isolation_scope() as scope:
            with capture_internal_exceptions():
                scope._name = "huey"
                scope.clear_breadcrumbs()
                scope.add_event_processor(_make_event_processor(task))

            sentry_headers = task.kwargs.pop("sentry_headers", None)
            is_span_streaming_enabled = has_span_streaming_enabled(
                sentry_sdk.get_client().options
            )

            if is_span_streaming_enabled:
                headers = sentry_headers or {}
                sentry_sdk.traces.continue_trace(headers)
                span_ctx = sentry_sdk.traces.start_span(
                    name=task.name,
                    attributes={
                        "sentry.op": OP.QUEUE_TASK_HUEY,
                        "sentry.origin": HueyIntegration.origin,
                        "sentry.span.source": SegmentSource.TASK,
                        SPANDATA.MESSAGING_DESTINATION_NAME: self.name,
                        "messaging.message.id": task.id,
                        "messaging.message.system": "huey",
                        "messaging.message.retry.count": (task.default_retries or 0)
                        - task.retries,
                    },
                    parent_span=None,
                )
            else:
                transaction = continue_trace(
                    sentry_headers or {},
                    name=task.name,
                    op=OP.QUEUE_TASK_HUEY,
                    source=TransactionSource.TASK,
                    origin=HueyIntegration.origin,
                )
                transaction.set_status(SPANSTATUS.OK)
                span_ctx = sentry_sdk.start_transaction(transaction)
                span_ctx.set_data(SPANDATA.MESSAGING_DESTINATION_NAME, self.name)

            if not getattr(task, "_sentry_is_patched", False):
                task.execute = _wrap_task_execute(task.execute)
                task._sentry_is_patched = True

            with span_ctx:
                return old_execute(self, task, timestamp)

    Huey._execute = _sentry_execute



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.utils import (
    capture_internal_exceptions,
    event_from_exception,
)

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any, Callable, Optional, Type

    Excepthook = Callable[
        [Type[BaseException], BaseException, Optional[TracebackType]],
        Any,
    ]

try:
    import typer
    from typer.main import except_hook
except ImportError:
    raise DidNotEnable("Typer not installed")


class TyperIntegration(Integration):
    identifier = "typer"

    @staticmethod
    def setup_once() -> None:
        typer.main.except_hook = _make_excepthook(except_hook)  # type: ignore


def _make_excepthook(old_excepthook: "Excepthook") -> "Excepthook":
    def sentry_sdk_excepthook(
        type_: "Type[BaseException]",
        value: BaseException,
        traceback: "Optional[TracebackType]",
    ) -> None:
        integration = sentry_sdk.get_client().get_integration(TyperIntegration)

        # Note: If we replace this with ensure_integration_enabled then
        # we break the exceptiongroup backport;
        # See: https://github.com/getsentry/sentry-python/issues/3097
        if integration is None:
            return old_excepthook(type_, value, traceback)

        with capture_internal_exceptions():
            event, hint = event_from_exception(
                (type_, value, traceback),
                client_options=sentry_sdk.get_client().options,
                mechanism={"type": "typer", "handled": False},
            )
            sentry_sdk.capture_event(event, hint=hint)

        return old_excepthook(type_, value, traceback)

    return sentry_sdk_excepthook



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import has_span_streaming_enabled

from ..consts import SPAN_ORIGIN
from ..utils import (
    _set_agent_data,
    _set_input_data,
    _set_output_data,
    _set_usage_data,
)

if TYPE_CHECKING:
    from typing import Any, Optional, Union

    from agents import Agent


def ai_client_span(
    agent: "Agent", get_response_kwargs: "dict[str, Any]"
) -> "Union[sentry_sdk.tracing.Span, StreamedSpan]":
    # TODO-anton: implement other types of operations. Now "chat" is hardcoded.
    # Get model name from agent.model or fall back to request model (for when agent.model is None/default)
    model_name = None
    if agent.model:
        model_name = agent.model.model if hasattr(agent.model, "model") else agent.model
    elif hasattr(agent, "_sentry_request_model"):
        model_name = agent._sentry_request_model

    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
    if span_streaming:
        span = sentry_sdk.traces.start_span(
            name=f"chat {model_name}",
            attributes={
                "sentry.op": OP.GEN_AI_CHAT,
                "sentry.origin": SPAN_ORIGIN,
                SPANDATA.GEN_AI_OPERATION_NAME: "chat",
            },
        )
    else:
        span = sentry_sdk.start_span(
            op=OP.GEN_AI_CHAT,
            name=f"chat {model_name}",
            origin=SPAN_ORIGIN,
        )
        # TODO-anton: remove hardcoded stuff and replace something that also works for embedding and so on
        span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "chat")

    _set_agent_data(span, agent)
    _set_input_data(span, get_response_kwargs)

    return span


def update_ai_client_span(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]",
    response: "Any",
    response_model: "Optional[str]" = None,
    agent: "Optional[Agent]" = None,
) -> None:
    """Update AI client span with response data (works for streaming and non-streaming)."""
    if hasattr(response, "usage") and response.usage:
        _set_usage_data(span, response.usage)

    if hasattr(response, "output") and response.output:
        _set_output_data(span, response)

    set_on_span = (
        span.set_attribute if isinstance(span, StreamedSpan) else span.set_data
    )

    if response_model is not None:
        set_on_span(SPANDATA.GEN_AI_RESPONSE_MODEL, response_model)
    elif hasattr(response, "model") and response.model:
        set_on_span(SPANDATA.GEN_AI_RESPONSE_MODEL, str(response.model))

    # Set conversation ID from agent if available
    if agent:
        conv_id = getattr(agent, "_sentry_conversation_id", None)
        if conv_id:
            set_on_span(SPANDATA.GEN_AI_CONVERSATION_ID, conv_id)



import os
import platform
import subprocess
import sys
from http.client import HTTPConnection, HTTPResponse
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import Integration
from sentry_sdk.scope import add_global_event_processor, should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing import Span
from sentry_sdk.tracing_utils import (
    EnvironHeaders,
    add_http_request_source,
    has_span_streaming_enabled,
    should_propagate_trace,
)
from sentry_sdk.utils import (
    SENSITIVE_DATA_SUBSTITUTE,
    capture_internal_exceptions,
    ensure_integration_enabled,
    is_sentry_url,
    logger,
    parse_url,
    safe_repr,
)

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, Optional, Union

    from sentry_sdk._types import Event, Hint


_RUNTIME_CONTEXT: "dict[str, object]" = {
    "name": platform.python_implementation(),
    "version": "%s.%s.%s" % (sys.version_info[:3]),
    "build": sys.version,
}


class StdlibIntegration(Integration):
    identifier = "stdlib"

    @staticmethod
    def setup_once() -> None:
        _install_httplib()
        _install_subprocess()

        @add_global_event_processor
        def add_python_runtime_context(
            event: "Event", hint: "Hint"
        ) -> "Optional[Event]":
            client = sentry_sdk.get_client()
            if client.get_integration(StdlibIntegration) is not None:
                contexts = event.setdefault("contexts", {})
                if isinstance(contexts, dict) and "runtime" not in contexts:
                    contexts["runtime"] = _RUNTIME_CONTEXT

            return event


def _complete_span(span: "Union[Span, StreamedSpan]") -> None:
    if isinstance(span, StreamedSpan):
        with capture_internal_exceptions():
            add_http_request_source(span)
        span.end()
    else:
        span.finish()
        with capture_internal_exceptions():
            add_http_request_source(span)


def _install_httplib() -> None:
    real_putrequest = HTTPConnection.putrequest
    real_getresponse = HTTPConnection.getresponse
    real_read = HTTPResponse.read
    real_close = HTTPResponse.close

    def putrequest(
        self: "HTTPConnection", method: str, url: str, *args: "Any", **kwargs: "Any"
    ) -> "Any":
        default_port = self.default_port

        # proxies go through set_tunnel
        tunnel_host = getattr(self, "_tunnel_host", None)
        if tunnel_host:
            host = tunnel_host
            port = getattr(self, "_tunnel_port", default_port)
        else:
            host = self.host
            port = self.port

        client = sentry_sdk.get_client()
        if client.get_integration(StdlibIntegration) is None or is_sentry_url(
            client, host
        ):
            return real_putrequest(self, method, url, *args, **kwargs)

        real_url = url
        if real_url is None or not real_url.startswith(("http://", "https://")):
            real_url = "%s://%s%s%s" % (
                default_port == 443 and "https" or "http",
                host,
                port != default_port and ":%s" % port or "",
                url,
            )

        parsed_url = None
        with capture_internal_exceptions():
            parsed_url = parse_url(real_url, sanitize=False)

        span_streaming = has_span_streaming_enabled(client.options)
        span: "Union[Span, StreamedSpan, None]"
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                span = None
            else:
                span = sentry_sdk.traces.start_span(
                    name="%s %s"
                    % (
                        method,
                        parsed_url.url if parsed_url else SENSITIVE_DATA_SUBSTITUTE,
                    ),
                    attributes={
                        "sentry.origin": "auto.http.stdlib.httplib",
                        "sentry.op": OP.HTTP_CLIENT,
                        SPANDATA.HTTP_REQUEST_METHOD: method,
                    },
                )

                if parsed_url is not None and should_send_default_pii():
                    span.set_attribute(SPANDATA.URL_FRAGMENT, parsed_url.fragment)
                    span.set_attribute(SPANDATA.URL_FULL, parsed_url.url)
                    span.set_attribute(SPANDATA.URL_QUERY, parsed_url.query)

                set_on_span = span.set_attribute
        else:
            span = sentry_sdk.start_span(
                op=OP.HTTP_CLIENT,
                name="%s %s"
                % (method, parsed_url.url if parsed_url else SENSITIVE_DATA_SUBSTITUTE),
                origin="auto.http.stdlib.httplib",
            )

            span.set_data(SPANDATA.HTTP_METHOD, method)
            if parsed_url is not None:
                span.set_data(SPANDATA.HTTP_FRAGMENT, parsed_url.fragment)
                span.set_data("url", parsed_url.url)
                span.set_data(SPANDATA.HTTP_QUERY, parsed_url.query)

            set_on_span = span.set_data

        # for proxies, these point to the proxy host/port
        if span and tunnel_host:
            set_on_span(SPANDATA.NETWORK_PEER_ADDRESS, self.host)
            set_on_span(SPANDATA.NETWORK_PEER_PORT, self.port)

        rv = real_putrequest(self, method, url, *args, **kwargs)

        if should_propagate_trace(client, real_url):
            for (
                key,
                value,
            ) in sentry_sdk.get_current_scope().iter_trace_propagation_headers(
                span=span
            ):
                logger.debug(
                    "[Tracing] Adding `{key}` header {value} to outgoing request to {real_url}.".format(
                        key=key, value=value, real_url=real_url
                    )
                )
                self.putheader(key, value)

        self._sentrysdk_span = span  # type: ignore[attr-defined]

        return rv

    def getresponse(self: "HTTPConnection", *args: "Any", **kwargs: "Any") -> "Any":
        span = getattr(self, "_sentrysdk_span", None)

        if span is None:
            return real_getresponse(self, *args, **kwargs)

        try:
            rv = real_getresponse(self, *args, **kwargs)
        except BaseException:
            _complete_span(span)
            raise

        if isinstance(span, StreamedSpan):
            status_code = int(rv.status)
            span.status = "error" if status_code >= 400 else "ok"
            span.set_attribute("http.response.status_code", status_code)
        else:
            span.set_http_status(int(rv.status))
            span.set_data("reason", rv.reason)

        # getresponse doesn't include actually reading the response body. This
        # is done in read(). So if the metadata/headers suggest there's a body to
        # read, don't finish the span just yet, but save it for ending it later.
        has_body = rv.chunked or (rv.length is not None and rv.length > 0)
        if has_body:
            rv._sentrysdk_span = span  # type: ignore[attr-defined]
        else:
            _complete_span(span)

        return rv

    def read(self: "HTTPResponse", *args: "Any", **kwargs: "Any") -> "Any":
        try:
            return real_read(self, *args, **kwargs)
        finally:
            span = getattr(self, "_sentrysdk_span", None)
            # read() might be called multiple times to consume a single body,
            # so we can't just end the span when read() is done. Instead,
            # try to figure out whether the response body has been fully read.
            if span and (self.fp is None or self.closed):
                self._sentrysdk_span = None  # type: ignore[attr-defined]
                _complete_span(span)

    def close(self: "HTTPResponse") -> None:
        # We patch close() as a best effort fallback in case the span is not
        # ended yet in getresponse() or read().

        try:
            real_close(self)
        finally:
            span = getattr(self, "_sentrysdk_span", None)
            if span is not None:
                self._sentrysdk_span = None  # type: ignore[attr-defined]
                _complete_span(span)

    HTTPConnection.putrequest = putrequest  # type: ignore[method-assign]
    HTTPConnection.getresponse = getresponse  # type: ignore[method-assign]
    HTTPResponse.read = read  # type: ignore[method-assign]
    HTTPResponse.close = close  # type: ignore[assignment,method-assign]


def _init_argument(
    args: "List[Any]",
    kwargs: "Dict[Any, Any]",
    name: str,
    position: int,
    setdefault_callback: "Optional[Callable[[Any], Any]]" = None,
) -> "Any":
    """
    given (*args, **kwargs) of a function call, retrieve (and optionally set a
    default for) an argument by either name or position.

    This is useful for wrapping functions with complex type signatures and
    extracting a few arguments without needing to redefine that function's
    entire type signature.
    """

    if name in kwargs:
        rv = kwargs[name]
        if setdefault_callback is not None:
            rv = setdefault_callback(rv)
        if rv is not None:
            kwargs[name] = rv
    elif position < len(args):
        rv = args[position]
        if setdefault_callback is not None:
            rv = setdefault_callback(rv)
        if rv is not None:
            args[position] = rv
    else:
        rv = setdefault_callback and setdefault_callback(None)
        if rv is not None:
            kwargs[name] = rv

    return rv


def _install_subprocess() -> None:
    old_popen_init = subprocess.Popen.__init__

    @ensure_integration_enabled(StdlibIntegration, old_popen_init)
    def sentry_patched_popen_init(
        self: "subprocess.Popen[Any]", *a: "Any", **kw: "Any"
    ) -> None:
        # Convert from tuple to list to be able to set values.
        a = list(a)

        args = _init_argument(a, kw, "args", 0) or []
        cwd = _init_argument(a, kw, "cwd", 9)

        # if args is not a list or tuple (and e.g. some iterator instead),
        # let's not use it at all. There are too many things that can go wrong
        # when trying to collect an iterator into a list and setting that list
        # into `a` again.
        #
        # Also invocations where `args` is not a sequence are not actually
        # legal. They just happen to work under CPython.
        description = None

        if isinstance(args, (list, tuple)) and len(args) < 100:
            with capture_internal_exceptions():
                description = " ".join(map(str, args))

        if description is None:
            description = safe_repr(args)

        env = None

        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        span: "Union[Span, StreamedSpan]"
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return old_popen_init(self, *a, **kw)

            span = sentry_sdk.traces.start_span(
                name=description,
                attributes={
                    "sentry.op": OP.SUBPROCESS,
                    "sentry.origin": "auto.subprocess.stdlib.subprocess",
                },
            )
        else:
            span = sentry_sdk.start_span(
                op=OP.SUBPROCESS,
                name=description,
                origin="auto.subprocess.stdlib.subprocess",
            )

        with span:
            for k, v in sentry_sdk.get_current_scope().iter_trace_propagation_headers(
                span=span
            ):
                if env is None:
                    env = _init_argument(
                        a,
                        kw,
                        "env",
                        10,
                        lambda x: dict(x if x is not None else os.environ),
                    )
                env["SUBPROCESS_" + k.upper().replace("-", "_")] = v

            if cwd and isinstance(span, Span):
                span.set_data("subprocess.cwd", cwd)

            rv = old_popen_init(self, *a, **kw)

            if isinstance(span, StreamedSpan):
                span.set_attribute(SPANDATA.PROCESS_PID, self.pid)
            else:
                span.set_tag("subprocess.pid", self.pid)

            return rv

    subprocess.Popen.__init__ = sentry_patched_popen_init  # type: ignore

    old_popen_wait = subprocess.Popen.wait

    @ensure_integration_enabled(StdlibIntegration, old_popen_wait)
    def sentry_patched_popen_wait(
        self: "subprocess.Popen[Any]", *a: "Any", **kw: "Any"
    ) -> "Any":
        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return old_popen_wait(self, *a, **kw)
            with sentry_sdk.traces.start_span(
                name=OP.SUBPROCESS_WAIT,
                attributes={
                    "sentry.op": OP.SUBPROCESS_WAIT,
                    "sentry.origin": "auto.subprocess.stdlib.subprocess",
                },
            ) as span:
                span.set_attribute(SPANDATA.PROCESS_PID, self.pid)
                return old_popen_wait(self, *a, **kw)
        else:
            with sentry_sdk.start_span(
                op=OP.SUBPROCESS_WAIT,
                origin="auto.subprocess.stdlib.subprocess",
            ) as span:
                span.set_tag("subprocess.pid", self.pid)
                return old_popen_wait(self, *a, **kw)

    subprocess.Popen.wait = sentry_patched_popen_wait  # type: ignore

    old_popen_communicate = subprocess.Popen.communicate

    @ensure_integration_enabled(StdlibIntegration, old_popen_communicate)
    def sentry_patched_popen_communicate(
        self: "subprocess.Popen[Any]", *a: "Any", **kw: "Any"
    ) -> "Any":
        span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
        if span_streaming:
            if sentry_sdk.traces.get_current_span() is None:
                return old_popen_communicate(self, *a, **kw)
            with sentry_sdk.traces.start_span(
                name=OP.SUBPROCESS_COMMUNICATE,
                attributes={
                    "sentry.op": OP.SUBPROCESS_COMMUNICATE,
                    "sentry.origin": "auto.subprocess.stdlib.subprocess",
                },
            ) as span:
                span.set_attribute(SPANDATA.PROCESS_PID, self.pid)
                return old_popen_communicate(self, *a, **kw)
        else:
            with sentry_sdk.start_span(
                op=OP.SUBPROCESS_COMMUNICATE,
                origin="auto.subprocess.stdlib.subprocess",
            ) as span:
                span.set_tag("subprocess.pid", self.pid)
                return old_popen_communicate(self, *a, **kw)

    subprocess.Popen.communicate = sentry_patched_popen_communicate  # type: ignore


def get_subprocess_traceparent_headers() -> "EnvironHeaders":
    return EnvironHeaders(os.environ, prefix="SUBPROCESS_")



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.integrations import Integration
from sentry_sdk.utils import capture_internal_exceptions, ensure_integration_enabled

if TYPE_CHECKING:
    from typing import Any, Optional

    from pyspark import SparkContext

    from sentry_sdk._types import Event, Hint


class SparkIntegration(Integration):
    identifier = "spark"

    @staticmethod
    def setup_once() -> None:
        _setup_sentry_tracing()


def _set_app_properties() -> None:
    """
    Set properties in driver that propagate to worker processes, allowing for workers to have access to those properties.
    This allows worker integration to have access to app_name and application_id.
    """
    from pyspark import SparkContext

    spark_context = SparkContext._active_spark_context
    if spark_context:
        spark_context.setLocalProperty(
            "sentry_app_name",
            spark_context.appName,
        )
        spark_context.setLocalProperty(
            "sentry_application_id",
            spark_context.applicationId,
        )


def _start_sentry_listener(sc: "SparkContext") -> None:
    """
    Start java gateway server to add custom `SparkListener`
    """
    from pyspark.java_gateway import ensure_callback_server_started

    gw = sc._gateway
    ensure_callback_server_started(gw)
    listener = SentryListener()
    sc._jsc.sc().addSparkListener(listener)


def _add_event_processor(sc: "SparkContext") -> None:
    scope = sentry_sdk.get_isolation_scope()

    @scope.add_event_processor
    def process_event(event: "Event", hint: "Hint") -> "Optional[Event]":
        with capture_internal_exceptions():
            if sentry_sdk.get_client().get_integration(SparkIntegration) is None:
                return event

            if sc._active_spark_context is None:
                return event

            event.setdefault("user", {}).setdefault("id", sc.sparkUser())

            event.setdefault("tags", {}).setdefault(
                "executor.id", sc._conf.get("spark.executor.id")
            )
            event["tags"].setdefault(
                "spark-submit.deployMode",
                sc._conf.get("spark.submit.deployMode"),
            )
            event["tags"].setdefault("driver.host", sc._conf.get("spark.driver.host"))
            event["tags"].setdefault("driver.port", sc._conf.get("spark.driver.port"))
            event["tags"].setdefault("spark_version", sc.version)
            event["tags"].setdefault("app_name", sc.appName)
            event["tags"].setdefault("application_id", sc.applicationId)
            event["tags"].setdefault("master", sc.master)
            event["tags"].setdefault("spark_home", sc.sparkHome)

            event.setdefault("extra", {}).setdefault("web_url", sc.uiWebUrl)

        return event


def _activate_integration(sc: "SparkContext") -> None:
    _start_sentry_listener(sc)
    _set_app_properties()
    _add_event_processor(sc)


def _patch_spark_context_init() -> None:
    from pyspark import SparkContext

    spark_context_init = SparkContext._do_init

    @ensure_integration_enabled(SparkIntegration, spark_context_init)
    def _sentry_patched_spark_context_init(
        self: "SparkContext", *args: "Any", **kwargs: "Any"
    ) -> "Optional[Any]":
        rv = spark_context_init(self, *args, **kwargs)
        _activate_integration(self)
        return rv

    SparkContext._do_init = _sentry_patched_spark_context_init


def _setup_sentry_tracing() -> None:
    from pyspark import SparkContext

    if SparkContext._active_spark_context is not None:
        _activate_integration(SparkContext._active_spark_context)
        return
    _patch_spark_context_init()


class SparkListener:
    def onApplicationEnd(self, applicationEnd: "Any") -> None:  # noqa: N802,N803
        pass

    def onApplicationStart(self, applicationStart: "Any") -> None:  # noqa: N802,N803
        pass

    def onBlockManagerAdded(self, blockManagerAdded: "Any") -> None:  # noqa: N802,N803
        pass

    def onBlockManagerRemoved(self, blockManagerRemoved: "Any") -> None:  # noqa: N802,N803
        pass

    def onBlockUpdated(self, blockUpdated: "Any") -> None:  # noqa: N802,N803
        pass

    def onEnvironmentUpdate(self, environmentUpdate: "Any") -> None:  # noqa: N802,N803
        pass

    def onExecutorAdded(self, executorAdded: "Any") -> None:  # noqa: N802,N803
        pass

    def onExecutorBlacklisted(self, executorBlacklisted: "Any") -> None:  # noqa: N802,N803
        pass

    def onExecutorBlacklistedForStage(  # noqa: N802
        self,
        executorBlacklistedForStage: "Any",  # noqa: N803
    ) -> None:
        pass

    def onExecutorMetricsUpdate(self, executorMetricsUpdate: "Any") -> None:  # noqa: N802,N803
        pass

    def onExecutorRemoved(self, executorRemoved: "Any") -> None:  # noqa: N802,N803
        pass

    def onJobEnd(self, jobEnd: "Any") -> None:  # noqa: N802,N803
        pass

    def onJobStart(self, jobStart: "Any") -> None:  # noqa: N802,N803
        pass

    def onNodeBlacklisted(self, nodeBlacklisted: "Any") -> None:  # noqa: N802,N803
        pass

    def onNodeBlacklistedForStage(self, nodeBlacklistedForStage: "Any") -> None:  # noqa: N802,N803
        pass

    def onNodeUnblacklisted(self, nodeUnblacklisted: "Any") -> None:  # noqa: N802,N803
        pass

    def onOtherEvent(self, event: "Any") -> None:  # noqa: N802,N803
        pass

    def onSpeculativeTaskSubmitted(self, speculativeTask: "Any") -> None:  # noqa: N802,N803
        pass

    def onStageCompleted(self, stageCompleted: "Any") -> None:  # noqa: N802,N803
        pass

    def onStageSubmitted(self, stageSubmitted: "Any") -> None:  # noqa: N802,N803
        pass

    def onTaskEnd(self, taskEnd: "Any") -> None:  # noqa: N802,N803
        pass

    def onTaskGettingResult(self, taskGettingResult: "Any") -> None:  # noqa: N802,N803
        pass

    def onTaskStart(self, taskStart: "Any") -> None:  # noqa: N802,N803
        pass

    def onUnpersistRDD(self, unpersistRDD: "Any") -> None:  # noqa: N802,N803
        pass

    class Java:
        implements = ["org.apache.spark.scheduler.SparkListenerInterface"]


class SentryListener(SparkListener):
    def _add_breadcrumb(
        self,
        level: str,
        message: str,
        data: "Optional[dict[str, Any]]" = None,
    ) -> None:
        sentry_sdk.get_isolation_scope().add_breadcrumb(
            level=level, message=message, data=data
        )

    def onJobStart(self, jobStart: "Any") -> None:  # noqa: N802,N803
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        message = "Job {} Started".format(jobStart.jobId())
        self._add_breadcrumb(level="info", message=message)
        _set_app_properties()

    def onJobEnd(self, jobEnd: "Any") -> None:  # noqa: N802,N803
        level = ""
        message = ""
        data = {"result": jobEnd.jobResult().toString()}

        if jobEnd.jobResult().toString() == "JobSucceeded":
            level = "info"
            message = "Job {} Ended".format(jobEnd.jobId())
        else:
            level = "warning"
            message = "Job {} Failed".format(jobEnd.jobId())

        self._add_breadcrumb(level=level, message=message, data=data)

    def onStageSubmitted(self, stageSubmitted: "Any") -> None:  # noqa: N802,N803
        stage_info = stageSubmitted.stageInfo()
        message = "Stage {} Submitted".format(stage_info.stageId())

        data = {"name": stage_info.name()}
        attempt_id = _get_attempt_id(stage_info)
        if attempt_id is not None:
            data["attemptId"] = attempt_id

        self._add_breadcrumb(level="info", message=message, data=data)
        _set_app_properties()

    def onStageCompleted(self, stageCompleted: "Any") -> None:  # noqa: N802,N803
        from py4j.protocol import Py4JJavaError  # type: ignore

        stage_info = stageCompleted.stageInfo()
        message = ""
        level = ""

        data = {"name": stage_info.name()}
        attempt_id = _get_attempt_id(stage_info)
        if attempt_id is not None:
            data["attemptId"] = attempt_id

        # Have to Try Except because stageInfo.failureReason() is typed with Scala Option
        try:
            data["reason"] = stage_info.failureReason().get()
            message = "Stage {} Failed".format(stage_info.stageId())
            level = "warning"
        except Py4JJavaError:
            message = "Stage {} Completed".format(stage_info.stageId())
            level = "info"

        self._add_breadcrumb(level=level, message=message, data=data)


def _get_attempt_id(stage_info: "Any") -> "Optional[int]":
    try:
        return stage_info.attemptId()
    except Exception:
        pass

    try:
        return stage_info.attemptNumber()
    except Exception:
        pass

    return None



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.integrations._wsgi_common import RequestExtractor
from sentry_sdk.integrations.wsgi import SentryWsgiMiddleware
from sentry_sdk.tracing import SOURCE_FOR_STYLE
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    parse_version,
)

if TYPE_CHECKING:
    from typing import Any, Dict, Optional

    from sentry_sdk._types import Event, EventProcessor

# In Falcon 3.0 `falcon.api_helpers` is renamed to `falcon.app_helpers`
# and `falcon.API` to `falcon.App`

try:
    import falcon  # type: ignore
    from falcon import __version__ as FALCON_VERSION
except ImportError:
    raise DidNotEnable("Falcon not installed")

try:
    import falcon.app_helpers  # type: ignore

    falcon_helpers = falcon.app_helpers
    falcon_app_class = falcon.App
    FALCON3 = True
except ImportError:
    import falcon.api_helpers  # type: ignore

    falcon_helpers = falcon.api_helpers
    falcon_app_class = falcon.API
    FALCON3 = False


_FALCON_UNSET: "Optional[object]" = None
if FALCON3:  # falcon.request._UNSET is only available in Falcon 3.0+
    with capture_internal_exceptions():
        from falcon.request import (  # type: ignore[import-not-found, no-redef]
            _UNSET as _FALCON_UNSET,
        )


class FalconRequestExtractor(RequestExtractor):
    def env(self) -> "Dict[str, Any]":
        return self.request.env

    def cookies(self) -> "Dict[str, Any]":
        return self.request.cookies

    def form(self) -> None:
        return None  # No such concept in Falcon

    def files(self) -> None:
        return None  # No such concept in Falcon

    def raw_data(self) -> "Optional[str]":
        # As request data can only be read once we won't make this available
        # to Sentry. Just send back a dummy string in case there was a
        # content length.
        # TODO(jmagnusson): Figure out if there's a way to support this
        content_length = self.content_length()
        if content_length > 0:
            return "[REQUEST_CONTAINING_RAW_DATA]"
        else:
            return None

    def json(self) -> "Optional[Dict[str, Any]]":
        # fallback to cached_media = None if self.request._media is not available
        cached_media = None
        with capture_internal_exceptions():
            # self.request._media is the cached self.request.media
            # value. It is only available if self.request.media
            # has already been accessed. Therefore, reading
            # self.request._media will not exhaust the raw request
            # stream (self.request.bounded_stream) because it has
            # already been read if self.request._media is set.
            cached_media = self.request._media

        if cached_media is not _FALCON_UNSET:
            return cached_media

        return None


class SentryFalconMiddleware:
    """Captures exceptions in Falcon requests and send to Sentry"""

    def process_request(
        self, req: "Any", resp: "Any", *args: "Any", **kwargs: "Any"
    ) -> None:
        integration = sentry_sdk.get_client().get_integration(FalconIntegration)
        if integration is None:
            return

        scope = sentry_sdk.get_isolation_scope()
        scope._name = "falcon"
        scope.add_event_processor(_make_request_event_processor(req, integration))

    def process_resource(
        self, req: "Any", resp: "Any", resource: "Any", params: "Any"
    ) -> None:
        """
        Sets the segment name and source as the route is resolved when this runs.
        """
        client = sentry_sdk.get_client()
        integration = client.get_integration(FalconIntegration)
        if integration is None or not has_span_streaming_enabled(client.options):
            return

        name_for_style = {
            "uri_template": req.uri_template,
            "path": req.path,
        }
        name = name_for_style[integration.transaction_style]
        source = sentry_sdk.traces.SOURCE_FOR_STYLE[integration.transaction_style]
        sentry_sdk.set_transaction_name(name, source)


TRANSACTION_STYLE_VALUES = ("uri_template", "path")


class FalconIntegration(Integration):
    identifier = "falcon"
    origin = f"auto.http.{identifier}"

    transaction_style = ""

    def __init__(self, transaction_style: str = "uri_template") -> None:
        if transaction_style not in TRANSACTION_STYLE_VALUES:
            raise ValueError(
                "Invalid value for transaction_style: %s (must be in %s)"
                % (transaction_style, TRANSACTION_STYLE_VALUES)
            )
        self.transaction_style = transaction_style

    @staticmethod
    def setup_once() -> None:
        version = parse_version(FALCON_VERSION)
        _check_minimum_version(FalconIntegration, version)

        _patch_wsgi_app()
        _patch_handle_exception()
        _patch_prepare_middleware()


def _patch_wsgi_app() -> None:
    original_wsgi_app = falcon_app_class.__call__

    def sentry_patched_wsgi_app(
        self: "falcon.API", env: "Any", start_response: "Any"
    ) -> "Any":
        integration = sentry_sdk.get_client().get_integration(FalconIntegration)
        if integration is None:
            return original_wsgi_app(self, env, start_response)

        sentry_wrapped = SentryWsgiMiddleware(
            lambda envi, start_resp: original_wsgi_app(self, envi, start_resp),
            span_origin=FalconIntegration.origin,
        )

        return sentry_wrapped(env, start_response)

    falcon_app_class.__call__ = sentry_patched_wsgi_app


def _patch_handle_exception() -> None:
    original_handle_exception = falcon_app_class._handle_exception

    @ensure_integration_enabled(FalconIntegration, original_handle_exception)
    def sentry_patched_handle_exception(self: "falcon.API", *args: "Any") -> "Any":
        # NOTE(jmagnusson): falcon 2.0 changed falcon.API._handle_exception
        # method signature from `(ex, req, resp, params)` to
        # `(req, resp, ex, params)`
        ex = response = None
        with capture_internal_exceptions():
            ex = next(argument for argument in args if isinstance(argument, Exception))
            response = next(
                argument for argument in args if isinstance(argument, falcon.Response)
            )

        was_handled = original_handle_exception(self, *args)

        if ex is None or response is None:
            # Both ex and response should have a non-None value at this point; otherwise,
            # there is an error with the SDK that will have been captured in the
            # capture_internal_exceptions block above.
            return was_handled

        if _exception_leads_to_http_5xx(ex, response):
            event, hint = event_from_exception(
                ex,
                client_options=sentry_sdk.get_client().options,
                mechanism={"type": "falcon", "handled": False},
            )
            sentry_sdk.capture_event(event, hint=hint)

        return was_handled

    falcon_app_class._handle_exception = sentry_patched_handle_exception


def _patch_prepare_middleware() -> None:
    original_prepare_middleware = falcon_helpers.prepare_middleware

    def sentry_patched_prepare_middleware(
        middleware: "Any" = None,
        independent_middleware: "Any" = False,
        asgi: bool = False,
    ) -> "Any":
        if asgi:
            # We don't support ASGI Falcon apps, so we don't patch anything here
            return original_prepare_middleware(middleware, independent_middleware, asgi)

        integration = sentry_sdk.get_client().get_integration(FalconIntegration)
        if integration is not None:
            middleware = [SentryFalconMiddleware()] + (middleware or [])

        # We intentionally omit the asgi argument here, since the default is False anyways,
        # and this way, we remain backwards-compatible with pre-3.0.0 Falcon versions.
        return original_prepare_middleware(middleware, independent_middleware)

    falcon_helpers.prepare_middleware = sentry_patched_prepare_middleware


def _exception_leads_to_http_5xx(ex: Exception, response: "falcon.Response") -> bool:
    is_server_error = isinstance(ex, falcon.HTTPError) and (ex.status or "").startswith(
        "5"
    )
    is_unhandled_error = not isinstance(
        ex, (falcon.HTTPError, falcon.http_status.HTTPStatus)
    )

    # We only check the HTTP status on Falcon 3 because in Falcon 2, the status on the response
    # at the stage where we capture it is listed as 200, even though we would expect to see a 500
    # status. Since at the time of this change, Falcon 2 is ca. 4 years old, we have decided to
    # only perform this check on Falcon 3+, despite the risk that some handled errors might be
    # reported to Sentry as unhandled on Falcon 2.
    return (is_server_error or is_unhandled_error) and (
        not FALCON3 or _has_http_5xx_status(response)
    )


def _has_http_5xx_status(response: "falcon.Response") -> bool:
    return response.status.startswith("5")


def _set_transaction_name_and_source(
    event: "Event", transaction_style: str, request: "falcon.Request"
) -> None:
    name_for_style = {
        "uri_template": request.uri_template,
        "path": request.path,
    }
    event["transaction"] = name_for_style[transaction_style]
    event["transaction_info"] = {"source": SOURCE_FOR_STYLE[transaction_style]}


def _make_request_event_processor(
    req: "falcon.Request", integration: "FalconIntegration"
) -> "EventProcessor":
    def event_processor(event: "Event", hint: "dict[str, Any]") -> "Event":
        _set_transaction_name_and_source(event, integration.transaction_style, req)

        with capture_internal_exceptions():
            FalconRequestExtractor(req).extract_into_event(event)

        return event

    return event_processor



import functools
import inspect
import sys

import sentry_sdk
from sentry_sdk.consts import OP, SPANSTATUS
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.traces import SegmentSource
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    event_from_exception,
    logger,
    package_version,
    qualname_from_function,
    reraise,
)

try:
    import ray  # type: ignore[import-not-found]
    from ray import remote
except ImportError:
    raise DidNotEnable("Ray not installed.")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Optional

    from sentry_sdk.utils import ExcInfo


def _check_sentry_initialized() -> None:
    if sentry_sdk.get_client().is_active():
        return

    logger.debug(
        "[Tracing] Sentry not initialized in ray cluster worker, performance data will be discarded."
    )


def _insert_sentry_tracing_in_signature(func: "Callable[..., Any]") -> None:
    # Patching new_func signature to add the _sentry_tracing parameter to it
    # Ray later inspects the signature and finds the unexpected parameter otherwise
    signature = inspect.signature(func)
    params = list(signature.parameters.values())
    sentry_tracing_param = inspect.Parameter(
        "_sentry_tracing",
        kind=inspect.Parameter.KEYWORD_ONLY,
        default=None,
    )

    # Keyword only arguments are penultimate if function has variadic keyword arguments
    if params and params[-1].kind is inspect.Parameter.VAR_KEYWORD:
        params.insert(-1, sentry_tracing_param)
    else:
        params.append(sentry_tracing_param)

    func.__signature__ = signature.replace(parameters=params)  # type: ignore[attr-defined]


def _patch_ray_remote() -> None:
    old_remote = remote

    @functools.wraps(old_remote)
    def new_remote(
        f: "Optional[Callable[..., Any]]" = None, *args: "Any", **kwargs: "Any"
    ) -> "Callable[..., Any]":
        if inspect.isclass(f):
            # Ray Actors
            # (https://docs.ray.io/en/latest/ray-core/actors.html)
            # are not supported
            # (Only Ray Tasks are supported)
            return old_remote(f, *args, **kwargs)

        def wrapper(user_f: "Callable[..., Any]") -> "Any":
            if inspect.isclass(user_f):
                # Ray Actors
                # (https://docs.ray.io/en/latest/ray-core/actors.html)
                # are not supported
                # (Only Ray Tasks are supported)
                return old_remote(*args, **kwargs)(user_f)

            @functools.wraps(user_f)
            def new_func(
                *f_args: "Any",
                _sentry_tracing: "Optional[dict[str, Any]]" = None,
                **f_kwargs: "Any",
            ) -> "Any":
                _check_sentry_initialized()

                span_streaming = has_span_streaming_enabled(
                    sentry_sdk.get_client().options
                )
                if span_streaming:
                    sentry_sdk.traces.continue_trace(_sentry_tracing or {})

                    function_name = qualname_from_function(user_f)
                    with sentry_sdk.traces.start_span(
                        name="unknown Ray task"
                        if function_name is None
                        else function_name,
                        attributes={
                            "sentry.op": OP.QUEUE_TASK_RAY,
                            "sentry.origin": RayIntegration.origin,
                            "sentry.span.source": SegmentSource.TASK,
                        },
                        parent_span=None,
                    ):
                        try:
                            result = user_f(*f_args, **f_kwargs)
                        except Exception:
                            exc_info = sys.exc_info()
                            _capture_exception(exc_info)
                            reraise(*exc_info)

                        return result
                else:
                    transaction = sentry_sdk.continue_trace(
                        _sentry_tracing or {},
                        op=OP.QUEUE_TASK_RAY,
                        name=qualname_from_function(user_f),
                        origin=RayIntegration.origin,
                        source=TransactionSource.TASK,
                    )

                    with sentry_sdk.start_transaction(transaction) as transaction:
                        try:
                            result = user_f(*f_args, **f_kwargs)
                            transaction.set_status(SPANSTATUS.OK)
                        except Exception:
                            transaction.set_status(SPANSTATUS.INTERNAL_ERROR)
                            exc_info = sys.exc_info()
                            _capture_exception(exc_info)
                            reraise(*exc_info)

                        return result

            _insert_sentry_tracing_in_signature(new_func)

            if f:
                rv = old_remote(new_func)
            else:
                rv = old_remote(*args, **kwargs)(new_func)
            old_remote_method = rv.remote

            def _remote_method_with_header_propagation(
                *args: "Any", **kwargs: "Any"
            ) -> "Any":
                """
                Ray Client
                """
                span_streaming = has_span_streaming_enabled(
                    sentry_sdk.get_client().options
                )
                if span_streaming:
                    function_name = qualname_from_function(user_f)

                    if sentry_sdk.traces.get_current_span() is None:
                        tracing = {
                            k: v
                            for k, v in sentry_sdk.get_current_scope().iter_trace_propagation_headers()
                        }
                        try:
                            result = old_remote_method(
                                *args, **kwargs, _sentry_tracing=tracing
                            )
                        except Exception:
                            exc_info = sys.exc_info()
                            _capture_exception(exc_info)
                            reraise(*exc_info)

                        return result

                    with sentry_sdk.traces.start_span(
                        name="unknown Ray task"
                        if function_name is None
                        else function_name,
                        attributes={
                            "sentry.op": OP.QUEUE_SUBMIT_RAY,
                            "sentry.origin": RayIntegration.origin,
                        },
                    ):
                        tracing = {
                            k: v
                            for k, v in sentry_sdk.get_current_scope().iter_trace_propagation_headers()
                        }
                        try:
                            result = old_remote_method(
                                *args, **kwargs, _sentry_tracing=tracing
                            )
                        except Exception:
                            exc_info = sys.exc_info()
                            _capture_exception(exc_info)
                            reraise(*exc_info)

                        return result
                else:
                    with sentry_sdk.start_span(
                        op=OP.QUEUE_SUBMIT_RAY,
                        name=qualname_from_function(user_f),
                        origin=RayIntegration.origin,
                    ) as span:
                        tracing = {
                            k: v
                            for k, v in sentry_sdk.get_current_scope().iter_trace_propagation_headers()
                        }
                        try:
                            result = old_remote_method(
                                *args, **kwargs, _sentry_tracing=tracing
                            )
                            span.set_status(SPANSTATUS.OK)
                        except Exception:
                            span.set_status(SPANSTATUS.INTERNAL_ERROR)
                            exc_info = sys.exc_info()
                            _capture_exception(exc_info)
                            reraise(*exc_info)

                        return result

            rv.remote = _remote_method_with_header_propagation

            return rv

        if f is not None:
            return wrapper(f)
        else:
            return wrapper

    ray.remote = new_remote


def _capture_exception(exc_info: "ExcInfo", **kwargs: "Any") -> None:
    client = sentry_sdk.get_client()

    event, hint = event_from_exception(
        exc_info,
        client_options=client.options,
        mechanism={
            "handled": False,
            "type": RayIntegration.identifier,
        },
    )
    sentry_sdk.capture_event(event, hint=hint)


class RayIntegration(Integration):
    identifier = "ray"
    origin = f"auto.queue.{identifier}"

    @staticmethod
    def setup_once() -> None:
        version = package_version("ray")
        _check_minimum_version(RayIntegration, version)

        _patch_ray_remote()



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Any, Awaitable, Callable, TypeVar

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import (
    add_query_source,
    has_span_streaming_enabled,
    record_sql_queries,
)
from sentry_sdk.utils import (
    capture_internal_exceptions,
    parse_version,
)

try:
    import aiomysql  # type: ignore[import-not-found]
    from aiomysql.connection import Connection  # type: ignore[import-not-found]
    from aiomysql.cursors import Cursor  # type: ignore[import-not-found]
except ImportError:
    raise DidNotEnable("aiomysql not installed.")


class AioMySQLIntegration(Integration):
    identifier = "aiomysql"
    origin = f"auto.db.{identifier}"
    _record_params = False

    def __init__(self, *, record_params: bool = False):
        AioMySQLIntegration._record_params = record_params

    @staticmethod
    def setup_once() -> None:
        aiomysql_version = parse_version(aiomysql.__version__)
        _check_minimum_version(AioMySQLIntegration, aiomysql_version)

        Cursor.execute = _wrap_execute(Cursor.execute)
        Cursor.executemany = _wrap_executemany(Cursor.executemany)

        # Patch Connection._connect — this catches ALL connections:
        #   - aiomysql.connect()
        #   - aiomysql.create_pool() (pool.py does `from .connection import connect`
        #     which ultimately calls Connection._connect)
        #   - Reconnects
        Connection._connect = _wrap_connect(Connection._connect)


T = TypeVar("T")


def _normalize_query(query: str | bytes | bytearray) -> str:
    if isinstance(query, (bytes, bytearray)):
        query = query.decode("utf-8", errors="replace")
    return " ".join(query.split())


def _wrap_execute(f: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
    """Wrap Cursor.execute to capture SQL queries."""

    async def _inner(*args: Any, **kwargs: Any) -> T:
        if sentry_sdk.get_client().get_integration(AioMySQLIntegration) is None:
            return await f(*args, **kwargs)

        cursor = args[0]

        # Skip if flagged by executemany (avoids double-recording).
        # Do NOT reset the flag here — it must stay True for the entire
        # duration of executemany, which may call execute multiple times
        # in a loop (non-INSERT fallback). Only _wrap_executemany's
        # finally block should clear it.
        if getattr(cursor, "_sentry_skip_next_execute", False):
            return await f(*args, **kwargs)

        query = args[1] if len(args) > 1 else kwargs.get("query", "")
        query_str = _normalize_query(query)
        params = args[2] if len(args) > 2 else kwargs.get("args")

        conn = _get_connection(cursor)

        integration = sentry_sdk.get_client().get_integration(AioMySQLIntegration)
        params_list = params if integration and integration._record_params else None
        param_style = "pyformat" if params_list else None

        with record_sql_queries(
            cursor=None,
            query=query_str,
            params_list=params_list,
            paramstyle=param_style,
            executemany=False,
            span_origin=AioMySQLIntegration.origin,
        ) as span:
            if conn:
                _set_db_data(span, conn)
            res = await f(*args, **kwargs)
            if isinstance(span, StreamedSpan):
                with capture_internal_exceptions():
                    add_query_source(span)

        if not isinstance(span, StreamedSpan):
            with capture_internal_exceptions():
                add_query_source(span)

        return res

    return _inner


def _wrap_executemany(f: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
    """Wrap Cursor.executemany to capture SQL queries."""

    async def _inner(*args: Any, **kwargs: Any) -> T:
        if sentry_sdk.get_client().get_integration(AioMySQLIntegration) is None:
            return await f(*args, **kwargs)

        cursor = args[0]
        query = args[1] if len(args) > 1 else kwargs.get("query", "")
        query_str = _normalize_query(query)
        seq_of_params = args[2] if len(args) > 2 else kwargs.get("args")

        conn = _get_connection(cursor)

        integration = sentry_sdk.get_client().get_integration(AioMySQLIntegration)
        params_list = (
            seq_of_params if integration and integration._record_params else None
        )
        param_style = "pyformat" if params_list else None

        # Prevent double-recording: _do_execute_many calls self.execute internally
        cursor._sentry_skip_next_execute = True
        try:
            with record_sql_queries(
                cursor=None,
                query=query_str,
                params_list=params_list,
                paramstyle=param_style,
                executemany=True,
                span_origin=AioMySQLIntegration.origin,
            ) as span:
                if conn:
                    _set_db_data(span, conn)
                res = await f(*args, **kwargs)
                if isinstance(span, StreamedSpan):
                    with capture_internal_exceptions():
                        add_query_source(span)

            if not isinstance(span, StreamedSpan):
                with capture_internal_exceptions():
                    add_query_source(span)

            return res
        finally:
            cursor._sentry_skip_next_execute = False

    return _inner


def _get_connection(cursor: Any) -> Any:
    """Get the underlying connection from a cursor."""
    return getattr(cursor, "connection", None)


def _wrap_connect(f: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
    """Wrap Connection._connect to capture connection spans."""

    async def _inner(self: "Connection") -> T:
        client = sentry_sdk.get_client()
        if client.get_integration(AioMySQLIntegration) is None:
            return await f(self)

        if has_span_streaming_enabled(client.options):
            breadcrumb_data = _get_connect_data(self, use_streaming_keys=True)

            with capture_internal_exceptions():
                sentry_sdk.add_breadcrumb(
                    message="connect", category="query", data=breadcrumb_data
                )

            if sentry_sdk.traces.get_current_span() is None:
                return await f(self)

            span_attributes: dict[str, Any] = {
                "sentry.op": OP.DB,
                "sentry.origin": AioMySQLIntegration.origin,
            } | breadcrumb_data

            with sentry_sdk.traces.start_span(
                name="connect", attributes=span_attributes
            ):
                return await f(self)

        connect_data = _get_connect_data(self)

        with sentry_sdk.start_span(
            op=OP.DB,
            name="connect",
            origin=AioMySQLIntegration.origin,
        ) as span:
            _set_db_data(span, self)

            with capture_internal_exceptions():
                sentry_sdk.add_breadcrumb(
                    message="connect",
                    category="query",
                    data=connect_data,
                )
            return await f(self)

    return _inner


def _get_connect_data(conn: Any, *, use_streaming_keys: bool = False) -> dict[str, Any]:
    if use_streaming_keys:
        db_system = SPANDATA.DB_SYSTEM_NAME
        db_name = SPANDATA.DB_NAMESPACE
    else:
        db_system = SPANDATA.DB_SYSTEM
        db_name = SPANDATA.DB_NAME

    data: dict[str, Any] = {
        db_system: "mysql",
        SPANDATA.DB_DRIVER_NAME: "aiomysql",
    }

    host = getattr(conn, "host", None)
    if host is not None:
        data[SPANDATA.SERVER_ADDRESS] = host

    port = getattr(conn, "port", None)
    if port is not None:
        data[SPANDATA.SERVER_PORT] = port

    database = getattr(conn, "db", None)
    if database is not None:
        data[db_name] = database

    user = getattr(conn, "user", None)
    if user is not None:
        data[SPANDATA.DB_USER] = user

    return data


def _set_db_data(span: Any, conn: Any) -> None:
    """Set database-related span data from connection object."""
    if isinstance(span, StreamedSpan):
        set_value = span.set_attribute
        db_system = SPANDATA.DB_SYSTEM_NAME
        db_name = SPANDATA.DB_NAMESPACE
    else:
        # Remove this else block once we've completely migrated to streamed spans
        # The use of deprecated attributes here is to ensure backwards compatibility
        set_value = span.set_data
        db_system = SPANDATA.DB_SYSTEM
        db_name = SPANDATA.DB_NAME

    set_value(db_system, "mysql")
    set_value(SPANDATA.DB_DRIVER_NAME, "aiomysql")

    host = getattr(conn, "host", None)
    if host is not None:
        set_value(SPANDATA.SERVER_ADDRESS, host)

    port = getattr(conn, "port", None)
    if port is not None:
        set_value(SPANDATA.SERVER_PORT, port)

    database = getattr(conn, "db", None)
    if database is not None:
        set_value(db_name, database)

    user = getattr(conn, "user", None)
    if user is not None:
        set_value(SPANDATA.DB_USER, user)



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.ai.utils import get_start_span_function
from sentry_sdk.tracing_utils import has_span_streaming_enabled

from ..consts import SPAN_ORIGIN

if TYPE_CHECKING:
    from typing import Union

    import agents


def agent_workflow_span(
    agent: "agents.Agent",
) -> "Union[sentry_sdk.tracing.Span, sentry_sdk.traces.StreamedSpan]":
    # Create a transaction or a span if an transaction is already active
    span_streaming = has_span_streaming_enabled(sentry_sdk.get_client().options)
    if span_streaming:
        span = sentry_sdk.traces.start_span(
            name=f"{agent.name} workflow", attributes={"sentry.origin": SPAN_ORIGIN}
        )

        return span

    span = get_start_span_function()(
        name=f"{agent.name} workflow",
        origin=SPAN_ORIGIN,
    )

    return span



import contextlib
import weakref
from inspect import iscoroutinefunction

import sentry_sdk
from sentry_sdk.api import continue_trace
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.integrations._wsgi_common import (
    RequestExtractor,
    _filter_headers,
    _is_json_content_type,
    request_body_within_bounds,
)
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import SegmentSource, StreamedSpan
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    CONTEXTVARS_ERROR_MESSAGE,
    HAS_REAL_CONTEXTVARS,
    AnnotatedValue,
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    transaction_from_function,
)

try:
    from tornado import version_info as TORNADO_VERSION
    from tornado.gen import coroutine
    from tornado.web import HTTPError, RequestHandler
except ImportError:
    raise DidNotEnable("Tornado not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, ContextManager, Dict, Generator, Optional, Union

    from sentry_sdk._types import Event, EventProcessor
    from sentry_sdk.tracing import Span


class TornadoIntegration(Integration):
    identifier = "tornado"
    origin = f"auto.http.{identifier}"

    @staticmethod
    def setup_once() -> None:
        _check_minimum_version(TornadoIntegration, TORNADO_VERSION)

        if not HAS_REAL_CONTEXTVARS:
            # Tornado is async. We better have contextvars or we're going to leak
            # state between requests.
            raise DidNotEnable(
                "The tornado integration for Sentry requires Python 3.7+ or the aiocontextvars package"
                + CONTEXTVARS_ERROR_MESSAGE
            )

        ignore_logger("tornado.access")

        old_execute = RequestHandler._execute

        awaitable = iscoroutinefunction(old_execute)

        if awaitable:
            # Starting Tornado 6 RequestHandler._execute method is a standard Python coroutine (async/await)
            # In that case our method should be a coroutine function too
            async def sentry_execute_request_handler(
                self: "RequestHandler", *args: "Any", **kwargs: "Any"
            ) -> "Any":
                with _handle_request_impl(self):
                    return await old_execute(self, *args, **kwargs)

        else:

            @coroutine  # type: ignore
            def sentry_execute_request_handler(
                self: "RequestHandler", *args: "Any", **kwargs: "Any"
            ) -> "Any":
                with _handle_request_impl(self):
                    result = yield from old_execute(self, *args, **kwargs)
                    return result

        RequestHandler._execute = sentry_execute_request_handler

        old_log_exception = RequestHandler.log_exception

        def sentry_log_exception(
            self: "Any",
            ty: type,
            value: BaseException,
            tb: "Any",
            *args: "Any",
            **kwargs: "Any",
        ) -> "Optional[Any]":
            _capture_exception(ty, value, tb)
            return old_log_exception(self, ty, value, tb, *args, **kwargs)

        RequestHandler.log_exception = sentry_log_exception


_DEFAULT_ROOT_SPAN_NAME = "generic Tornado request"


@contextlib.contextmanager
def _handle_request_impl(self: "RequestHandler") -> "Generator[None, None, None]":
    integration = sentry_sdk.get_client().get_integration(TornadoIntegration)

    if integration is None:
        yield
        return

    weak_handler = weakref.ref(self)
    client = sentry_sdk.get_client()
    is_span_streaming_enabled = has_span_streaming_enabled(client.options)

    with sentry_sdk.isolation_scope() as scope:
        headers = self.request.headers

        scope.clear_breadcrumbs()
        processor = _make_event_processor(weak_handler)
        scope.add_event_processor(processor)

        span_ctx: "ContextManager[Union[Span, StreamedSpan, None]]"

        if is_span_streaming_enabled:
            sentry_sdk.traces.continue_trace(dict(headers))
            scope.set_custom_sampling_context({"tornado_request": self.request})

            if should_send_default_pii() and self.request.remote_ip:
                scope.set_attribute(SPANDATA.USER_IP_ADDRESS, self.request.remote_ip)

            span_ctx = sentry_sdk.traces.start_span(
                name=_DEFAULT_ROOT_SPAN_NAME,
                attributes={
                    "sentry.op": OP.HTTP_SERVER,
                    "sentry.origin": TornadoIntegration.origin,
                    "sentry.span.source": SegmentSource.ROUTE,
                },
                parent_span=None,
            )
        else:
            transaction = continue_trace(
                headers,
                op=OP.HTTP_SERVER,
                # Like with all other integrations, this is our
                # fallback transaction in case there is no route.
                # sentry_urldispatcher_resolve is responsible for
                # setting a transaction name later.
                name=_DEFAULT_ROOT_SPAN_NAME,
                source=TransactionSource.ROUTE,
                origin=TornadoIntegration.origin,
            )
            span_ctx = sentry_sdk.start_transaction(
                transaction,
                custom_sampling_context={"tornado_request": self.request},
            )

        with span_ctx as span:
            try:
                yield
            finally:
                if type(span) is StreamedSpan:
                    with capture_internal_exceptions():
                        for attr, value in _get_request_attributes(
                            self.request
                        ).items():
                            span.set_attribute(attr, value)

                    with capture_internal_exceptions():
                        method = getattr(self, self.request.method.lower(), None)
                        if method is not None:
                            span_name = transaction_from_function(method)
                            if span_name:
                                span.name = span_name
                                span.set_attribute(
                                    "sentry.span.source",
                                    SegmentSource.COMPONENT,
                                )

                    with capture_internal_exceptions():
                        status_int = self.get_status()
                        span.set_attribute(SPANDATA.HTTP_STATUS_CODE, status_int)
                        span.status = "error" if status_int >= 400 else "ok"


def _get_request_attributes(request: "Any") -> "Dict[str, Any]":
    attributes = {}  # type: Dict[str, Any]

    if request.method:
        attributes[SPANDATA.HTTP_REQUEST_METHOD] = request.method.upper()

    headers = _filter_headers(dict(request.headers), use_annotated_value=False)
    for header, value in headers.items():
        attributes[f"{SPANDATA.HTTP_REQUEST_HEADER}.{header.lower()}"] = value

    if should_send_default_pii():
        attributes[SPANDATA.URL_FULL] = request.full_url()
        attributes["url.path"] = request.path

        if request.query:
            attributes[SPANDATA.URL_QUERY] = request.query

    if request.protocol:
        attributes[SPANDATA.NETWORK_PROTOCOL_NAME] = request.protocol

    if should_send_default_pii() and request.remote_ip:
        attributes[SPANDATA.CLIENT_ADDRESS] = request.remote_ip

    with capture_internal_exceptions():
        raw_data = _get_tornado_request_data(request)
        body_data = raw_data.value if isinstance(raw_data, AnnotatedValue) else raw_data
        if body_data is not None:
            attributes[SPANDATA.HTTP_REQUEST_BODY_DATA] = body_data

    return attributes


def _get_tornado_request_data(
    request: "Any",
) -> "Union[Optional[str], AnnotatedValue]":
    body = request.body
    if not body:
        return None

    if not request_body_within_bounds(sentry_sdk.get_client(), len(body)):
        return AnnotatedValue.substituted_because_over_size_limit()

    return body.decode("utf-8", "replace")


@ensure_integration_enabled(TornadoIntegration)
def _capture_exception(ty: type, value: BaseException, tb: "Any") -> None:
    if isinstance(value, HTTPError):
        return

    event, hint = event_from_exception(
        (ty, value, tb),
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": "tornado", "handled": False},
    )

    sentry_sdk.capture_event(event, hint=hint)


def _make_event_processor(
    weak_handler: "Callable[[], RequestHandler]",
) -> "EventProcessor":
    def tornado_processor(event: "Event", hint: "dict[str, Any]") -> "Event":
        handler = weak_handler()
        if handler is None:
            return event

        request = handler.request

        with capture_internal_exceptions():
            method = getattr(handler, handler.request.method.lower())
            event["transaction"] = transaction_from_function(method) or ""
            event["transaction_info"] = {"source": TransactionSource.COMPONENT}

        with capture_internal_exceptions():
            extractor = TornadoRequestExtractor(request)
            extractor.extract_into_event(event)

            request_info = event["request"]

            request_info["url"] = "%s://%s%s" % (
                request.protocol,
                request.host,
                request.path,
            )

            request_info["query_string"] = request.query
            request_info["method"] = request.method
            request_info["env"] = {"REMOTE_ADDR": request.remote_ip}
            request_info["headers"] = _filter_headers(dict(request.headers))

        if should_send_default_pii():
            try:
                current_user = handler.current_user
            except Exception:
                current_user = None

            if current_user:
                event.setdefault("user", {}).setdefault("is_authenticated", True)

        return event

    return tornado_processor


class TornadoRequestExtractor(RequestExtractor):
    def content_length(self) -> int:
        if self.request.body is None:
            return 0
        return len(self.request.body)

    def cookies(self) -> "Dict[str, str]":
        return {k: v.value for k, v in self.request.cookies.items()}

    def raw_data(self) -> bytes:
        return self.request.body

    def form(self) -> "Dict[str, Any]":
        return {
            k: [v.decode("latin1", "replace") for v in vs]
            for k, vs in self.request.body_arguments.items()
        }

    def is_json(self) -> bool:
        return _is_json_content_type(self.request.headers.get("content-type"))

    def files(self) -> "Dict[str, Any]":
        return {k: v[0] for k, v in self.request.files.items() if v}

    def size_of_file(self, file: "Any") -> int:
        return len(file.body or ())



import time
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Tuple

    from sentry_sdk._types import MonitorConfigScheduleUnit


def _now_seconds_since_epoch() -> float:
    # We cannot use `time.perf_counter()` when dealing with the duration
    # of a Celery task, because the start of a Celery task and
    # the end are recorded in different processes.
    # Start happens in the Celery Beat process,
    # the end in a Celery Worker process.
    return time.time()


def _get_humanized_interval(seconds: float) -> "Tuple[int, MonitorConfigScheduleUnit]":
    TIME_UNITS = (  # noqa: N806
        ("day", 60 * 60 * 24.0),
        ("hour", 60 * 60.0),
        ("minute", 60.0),
    )

    seconds = float(seconds)
    for unit, divider in TIME_UNITS:
        if seconds >= divider:
            interval = int(seconds / divider)
            return (interval, cast("MonitorConfigScheduleUnit", unit))

    return (int(seconds), "second")



from functools import wraps
from typing import Any

from sentry_sdk.feature_flags import add_feature_flag
from sentry_sdk.integrations import DidNotEnable, Integration

try:
    from UnleashClient import UnleashClient
except ImportError:
    raise DidNotEnable("UnleashClient is not installed")


class UnleashIntegration(Integration):
    identifier = "unleash"

    @staticmethod
    def setup_once() -> None:
        # Wrap and patch evaluation methods (class methods)
        old_is_enabled = UnleashClient.is_enabled

        @wraps(old_is_enabled)
        def sentry_is_enabled(
            self: "UnleashClient", feature: str, *args: "Any", **kwargs: "Any"
        ) -> "Any":
            enabled = old_is_enabled(self, feature, *args, **kwargs)

            # We have no way of knowing what type of unleash feature this is, so we have to treat
            # it as a boolean / toggle feature.
            add_feature_flag(feature, enabled)

            return enabled

        UnleashClient.is_enabled = sentry_is_enabled  # type: ignore



import inspect
import sys
from functools import wraps
from typing import TYPE_CHECKING

import sentry_sdk.utils
from sentry_sdk import start_span
from sentry_sdk.ai.utils import _set_span_data_attribute
from sentry_sdk.consts import SPANDATA
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing import Span
from sentry_sdk.utils import ContextVar, capture_internal_exceptions, reraise

if TYPE_CHECKING:
    from typing import Any, Awaitable, Callable, Optional, TypeVar, Union

    F = TypeVar("F", bound=Union[Callable[..., Any], Callable[..., Awaitable[Any]]])

_ai_pipeline_name = ContextVar("ai_pipeline_name", default=None)


def set_ai_pipeline_name(name: "Optional[str]") -> None:
    _ai_pipeline_name.set(name)


def get_ai_pipeline_name() -> "Optional[str]":
    return _ai_pipeline_name.get()


def ai_track(description: str, **span_kwargs: "Any") -> "Callable[[F], F]":
    def decorator(f: "F") -> "F":
        def sync_wrapped(*args: "Any", **kwargs: "Any") -> "Any":
            curr_pipeline = _ai_pipeline_name.get()
            op = span_kwargs.pop("op", "ai.run" if curr_pipeline else "ai.pipeline")

            with start_span(name=description, op=op, **span_kwargs) as span:
                for k, v in kwargs.pop("sentry_tags", {}).items():
                    span.set_tag(k, v)
                for k, v in kwargs.pop("sentry_data", {}).items():
                    span.set_data(k, v)
                if curr_pipeline:
                    span.set_data(SPANDATA.GEN_AI_PIPELINE_NAME, curr_pipeline)
                    return f(*args, **kwargs)
                else:
                    _ai_pipeline_name.set(description)
                    try:
                        res = f(*args, **kwargs)
                    except Exception as e:
                        exc_info = sys.exc_info()
                        with capture_internal_exceptions():
                            event, hint = sentry_sdk.utils.event_from_exception(
                                e,
                                client_options=sentry_sdk.get_client().options,
                                mechanism={"type": "ai_monitoring", "handled": False},
                            )
                            sentry_sdk.capture_event(event, hint=hint)
                        reraise(*exc_info)
                    finally:
                        _ai_pipeline_name.set(None)
                    return res

        async def async_wrapped(*args: "Any", **kwargs: "Any") -> "Any":
            curr_pipeline = _ai_pipeline_name.get()
            op = span_kwargs.pop("op", "ai.run" if curr_pipeline else "ai.pipeline")

            with start_span(name=description, op=op, **span_kwargs) as span:
                for k, v in kwargs.pop("sentry_tags", {}).items():
                    span.set_tag(k, v)
                for k, v in kwargs.pop("sentry_data", {}).items():
                    span.set_data(k, v)
                if curr_pipeline:
                    span.set_data(SPANDATA.GEN_AI_PIPELINE_NAME, curr_pipeline)
                    return await f(*args, **kwargs)
                else:
                    _ai_pipeline_name.set(description)
                    try:
                        res = await f(*args, **kwargs)
                    except Exception as e:
                        exc_info = sys.exc_info()
                        with capture_internal_exceptions():
                            event, hint = sentry_sdk.utils.event_from_exception(
                                e,
                                client_options=sentry_sdk.get_client().options,
                                mechanism={"type": "ai_monitoring", "handled": False},
                            )
                            sentry_sdk.capture_event(event, hint=hint)
                        reraise(*exc_info)
                    finally:
                        _ai_pipeline_name.set(None)
                    return res

        if inspect.iscoroutinefunction(f):
            return wraps(f)(async_wrapped)  # type: ignore
        else:
            return wraps(f)(sync_wrapped)  # type: ignore

    return decorator


def record_token_usage(
    span: "Union[Span, StreamedSpan]",
    input_tokens: "Optional[int]" = None,
    input_tokens_cached: "Optional[int]" = None,
    input_tokens_cache_write: "Optional[int]" = None,
    output_tokens: "Optional[int]" = None,
    output_tokens_reasoning: "Optional[int]" = None,
    total_tokens: "Optional[int]" = None,
) -> None:
    # TODO: move pipeline name elsewhere
    ai_pipeline_name = get_ai_pipeline_name()
    if ai_pipeline_name:
        _set_span_data_attribute(span, SPANDATA.GEN_AI_PIPELINE_NAME, ai_pipeline_name)

    if input_tokens is not None:
        _set_span_data_attribute(span, SPANDATA.GEN_AI_USAGE_INPUT_TOKENS, input_tokens)

    if input_tokens_cached is not None:
        _set_span_data_attribute(
            span,
            SPANDATA.GEN_AI_USAGE_INPUT_TOKENS_CACHED,
            input_tokens_cached,
        )

    if input_tokens_cache_write is not None:
        _set_span_data_attribute(
            span,
            SPANDATA.GEN_AI_USAGE_INPUT_TOKENS_CACHE_WRITE,
            input_tokens_cache_write,
        )

    if output_tokens is not None:
        _set_span_data_attribute(
            span, SPANDATA.GEN_AI_USAGE_OUTPUT_TOKENS, output_tokens
        )

    if output_tokens_reasoning is not None:
        _set_span_data_attribute(
            span,
            SPANDATA.GEN_AI_USAGE_OUTPUT_TOKENS_REASONING,
            output_tokens_reasoning,
        )

    if total_tokens is None and input_tokens is not None and output_tokens is not None:
        total_tokens = input_tokens + output_tokens

    if total_tokens is not None:
        _set_span_data_attribute(span, SPANDATA.GEN_AI_USAGE_TOTAL_TOKENS, total_tokens)



"""
Code used for the Caches module in Sentry
"""

from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations.redis.utils import _get_safe_key, _key_as_string
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.utils import capture_internal_exceptions

GET_COMMANDS = ("get", "mget")
SET_COMMANDS = ("set", "setex")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Optional, Union

    from sentry_sdk.integrations.redis import RedisIntegration
    from sentry_sdk.tracing import Span


def _get_op(name: str) -> "Optional[str]":
    op = None
    if name.lower() in GET_COMMANDS:
        op = OP.CACHE_GET
    elif name.lower() in SET_COMMANDS:
        op = OP.CACHE_PUT

    return op


def _compile_cache_span_properties(
    redis_command: str,
    args: "tuple[Any, ...]",
    kwargs: "dict[str, Any]",
    integration: "RedisIntegration",
) -> "dict[str, Any]":
    key = _get_safe_key(redis_command, args, kwargs)
    key_as_string = _key_as_string(key)
    keys_as_string = key_as_string.split(", ")

    is_cache_key = False
    for prefix in integration.cache_prefixes:
        for kee in keys_as_string:
            if kee.startswith(prefix):
                is_cache_key = True
                break
        if is_cache_key:
            break

    value = None
    if redis_command.lower() in SET_COMMANDS:
        value = args[-1]

    properties = {
        "op": _get_op(redis_command),
        "description": _get_cache_span_description(
            redis_command, args, kwargs, integration
        ),
        "key": key,
        "key_as_string": key_as_string,
        "redis_command": redis_command.lower(),
        "is_cache_key": is_cache_key,
        "value": value,
    }

    return properties


def _get_cache_span_description(
    redis_command: str,
    args: "tuple[Any, ...]",
    kwargs: "dict[str, Any]",
    integration: "RedisIntegration",
) -> str:
    description = _key_as_string(_get_safe_key(redis_command, args, kwargs))

    if integration.max_data_size and len(description) > integration.max_data_size:
        description = description[: integration.max_data_size - len("...")] + "..."

    return description


def _set_cache_data(
    span: "Union[Span, StreamedSpan]",
    redis_client: "Any",
    properties: "dict[str, Any]",
    return_value: "Optional[Any]",
) -> None:
    if isinstance(span, StreamedSpan):
        set_on_span = span.set_attribute
    else:
        set_on_span = span.set_data

    with capture_internal_exceptions():
        set_on_span(SPANDATA.CACHE_KEY, properties["key"])

        if properties["redis_command"] in GET_COMMANDS:
            if return_value is not None:
                set_on_span(SPANDATA.CACHE_HIT, True)
                size = (
                    len(str(return_value).encode("utf-8"))
                    if not isinstance(return_value, bytes)
                    else len(return_value)
                )
                set_on_span(SPANDATA.CACHE_ITEM_SIZE, size)
            else:
                set_on_span(SPANDATA.CACHE_HIT, False)

        elif properties["redis_command"] in SET_COMMANDS:
            if properties["value"] is not None:
                size = (
                    len(properties["value"].encode("utf-8"))
                    if not isinstance(properties["value"], bytes)
                    else len(properties["value"])
                )
                set_on_span(SPANDATA.CACHE_ITEM_SIZE, size)

        try:
            connection_params = redis_client.connection_pool.connection_kwargs
        except AttributeError:
            # If it is a cluster, there is no connection_pool attribute so we
            # need to get the default node from the cluster instance
            default_node = redis_client.get_default_node()
            connection_params = {
                "host": default_node.host,
                "port": default_node.port,
            }

        host = connection_params.get("host")
        if host is not None:
            set_on_span(SPANDATA.NETWORK_PEER_ADDRESS, host)

        port = connection_params.get("port")
        if port is not None:
            set_on_span(SPANDATA.NETWORK_PEER_PORT, port)



from sentry_sdk import metrics, profiler

from sentry_sdk.scope import Scope  # isort: skip
from sentry_sdk.client import Client  # isort: skip
from sentry_sdk.consts import VERSION
from sentry_sdk.transport import HttpTransport, Transport

from sentry_sdk.api import *  # noqa # isort: skip

__all__ = [  # noqa
    "Hub",
    "Scope",
    "Client",
    "Transport",
    "HttpTransport",
    "VERSION",
    "integrations",
    # From sentry_sdk.api
    "init",
    "add_attachment",
    "add_breadcrumb",
    "capture_event",
    "capture_exception",
    "capture_message",
    "configure_scope",
    "continue_trace",
    "flush",
    "flush_async",
    "get_baggage",
    "get_client",
    "get_global_scope",
    "get_isolation_scope",
    "get_current_scope",
    "get_current_span",
    "get_traceparent",
    "is_initialized",
    "isolation_scope",
    "last_event_id",
    "new_scope",
    "push_scope",
    "remove_attribute",
    "set_attribute",
    "set_context",
    "set_extra",
    "set_level",
    "set_measurement",
    "set_tag",
    "set_tags",
    "set_user",
    "start_span",
    "start_transaction",
    "trace",
    "monitor",
    "logger",
    "metrics",
    "profiler",
    "start_session",
    "end_session",
    "set_transaction_name",
    "update_current_span",
]

# Initialize the debug support after everything is loaded
from sentry_sdk.debug import init_debug_support

init_debug_support()
del init_debug_support

# circular imports
from sentry_sdk.hub import Hub



import sys
from copy import deepcopy
from functools import wraps
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan, get_current_span
from sentry_sdk.tracing import SOURCE_FOR_STYLE, TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import transaction_from_function

if TYPE_CHECKING:
    from typing import Any, Awaitable, Callable, Dict

    from sentry_sdk._types import Event

try:
    from sentry_sdk.integrations.starlette import (
        StarletteIntegration,
        StarletteRequestExtractor,
        _get_cached_request_body_attribute,
    )
except DidNotEnable:
    raise DidNotEnable("Starlette is not installed")

try:
    import fastapi  # type: ignore
except ImportError:
    raise DidNotEnable("FastAPI is not installed")


_DEFAULT_TRANSACTION_NAME = "generic FastAPI request"


# Vendored: https://github.com/Kludex/starlette/blob/0a29b5ccdcbd1285c75c4fdb5d62ae1d244a21b0/starlette/_utils.py#L11-L17
if sys.version_info >= (3, 13):  # pragma: no cover
    from inspect import iscoroutinefunction
else:
    from asyncio import iscoroutinefunction


class FastApiIntegration(StarletteIntegration):
    identifier = "fastapi"

    @staticmethod
    def setup_once() -> None:
        patch_get_request_handler()


def _set_transaction_name_and_source(
    scope: "sentry_sdk.Scope", transaction_style: str, request: "Any"
) -> None:
    name = ""

    if transaction_style == "endpoint":
        endpoint = request.scope.get("endpoint")
        if endpoint:
            name = transaction_from_function(endpoint) or ""

    elif transaction_style == "url":
        route = request.scope.get("route")

        if route:
            # FastAPI >= 0.137 stores the prefix-resolved path on an
            # effective_route_context in scope["fastapi"], while
            # scope["route"].path holds the unprefixed original.
            # Prefer the effective context path when available.
            effective_route_context = request.scope.get("fastapi", {}).get(
                "effective_route_context"
            )
            context_path = getattr(effective_route_context, "path", None)

            if context_path:
                name = context_path
            else:
                path = getattr(route, "path", None)
                if path is not None:
                    name = path

    if not name:
        name = _DEFAULT_TRANSACTION_NAME
        source = TransactionSource.ROUTE
    else:
        source = SOURCE_FOR_STYLE[transaction_style]

    scope.set_transaction_name(name, source=source)


async def _wrap_async_handler(
    handler: "Callable[..., Awaitable[Any]]", *args: "Any", **kwargs: "Any"
) -> "Any":
    """
    Wraps an asynchronous handler function to attach request info to errors and the server segment span.
    The request body cached on the Starlette Request object is attached to streamed spans, but consuming the request body in the event
    processor can still cause application hangs.
    """
    client = sentry_sdk.get_client()
    integration = client.get_integration(FastApiIntegration)
    if integration is None:
        return await handler(*args, **kwargs)

    request = args[0]

    _set_transaction_name_and_source(
        sentry_sdk.get_current_scope(), integration.transaction_style, request
    )
    sentry_scope = sentry_sdk.get_isolation_scope()
    extractor = StarletteRequestExtractor(request)
    info = await extractor.extract_request_info()

    def _make_request_event_processor(
        req: "Any", integration: "Any"
    ) -> "Callable[[Event, Dict[str, Any]], Event]":
        def event_processor(event: "Event", hint: "Dict[str, Any]") -> "Event":
            # Extract information from request
            request_info = event.get("request", {})
            if info:
                if "cookies" in info and should_send_default_pii():
                    request_info["cookies"] = info["cookies"]
                if "data" in info:
                    request_info["data"] = info["data"]
            event["request"] = deepcopy(request_info)

            return event

        return event_processor

    sentry_scope._name = FastApiIntegration.identifier
    sentry_scope.add_event_processor(
        _make_request_event_processor(request, integration)
    )

    try:
        return await handler(*args, **kwargs)
    finally:
        current_span = get_current_span()

        if type(current_span) is StreamedSpan:
            request_body = _get_cached_request_body_attribute(
                client=client, request=request
            )
            if request_body:
                current_span._segment.set_attribute(
                    SPANDATA.HTTP_REQUEST_BODY_DATA,
                    request_body,
                )


def patch_get_request_handler() -> None:
    old_get_request_handler = fastapi.routing.get_request_handler

    def _sentry_get_request_handler(*args: "Any", **kwargs: "Any") -> "Any":
        dependant = kwargs.get("dependant")
        if (
            dependant
            and dependant.call is not None
            and not iscoroutinefunction(dependant.call)
            # FastAPI >= 0.137 calls get_request_handler() on every request
            # (router-tree traversal) rather than once at registration. Guard
            # against accumulating _sentry_call wrappers on the shared
            # dependant object, which would cause a RecursionError after ~987
            # requests as the call chain grows past Python's recursion limit.
            and not getattr(dependant.call, "_sentry_is_patched", False)
        ):
            old_call = dependant.call

            @wraps(old_call)
            def _sentry_call(*args: "Any", **kwargs: "Any") -> "Any":
                current_scope = sentry_sdk.get_current_scope()

                client = sentry_sdk.get_client()
                if has_span_streaming_enabled(client.options):
                    current_span = current_scope.streamed_span

                    if type(current_span) is StreamedSpan:
                        segment = current_span._segment
                        segment._update_active_thread()

                elif current_scope.transaction is not None:
                    current_scope.transaction.update_active_thread()

                sentry_scope = sentry_sdk.get_isolation_scope()
                if sentry_scope.profile is not None:
                    sentry_scope.profile.update_active_thread_id()

                return old_call(*args, **kwargs)

            _sentry_call._sentry_is_patched = True  # type: ignore[attr-defined]
            dependant.call = _sentry_call

        old_app = old_get_request_handler(*args, **kwargs)

        async def _sentry_app(*args: "Any", **kwargs: "Any") -> "Any":
            return await _wrap_async_handler(old_app, *args, **kwargs)

        return _sentry_app

    fastapi.routing.get_request_handler = _sentry_get_request_handler
